# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
from collections import defaultdict
from copy import deepcopy
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generator,
    KeysView,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    ValuesView,
    cast,
)

import networkx as nx  # type:ignore
import networkx.algorithms.isomorphism as iso  # type:ignore
from networkx.classes.reportviews import OutEdgeView  # type:ignore

import nncf
from nncf.common.graph.graph_matching import find_subgraphs_matching_pattern
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OUTPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.patterns import GraphPattern
from nncf.common.utils.dot_file_rw import write_dot_graph

NNCFNodeName = str
LayerName = str


class NNCFNode:
    """
    Class describing nodes used in NNCFGraph.
    """

    ID_NODE_ATTR = "id"
    NODE_NAME_ATTR = "node_name"
    KEY_NODE_ATTR = "key"
    NODE_TYPE_ATTR = "type"
    METATYPE_ATTR = "metatype"
    LAYER_NAME_ATTR = "layer_name"
    LAYER_ATTRIBUTES = "layer_attributes"
    IGNORED_ALGOS_ATTR = "ignored_algos"
    IS_IN_ITERATION_SCOPE_NODE_ATTR = "is_in_iteration_scope"
    IS_INTEGER_INPUT_NODE_ATTR = "is_integer_input"
    IS_SHARED_ATTR = "is_shared"

    def __init__(self, attributes: Dict[str, Any]) -> None:
        self._attributes = attributes

    @property
    def attributes(self) -> Dict[str, Any]:
        return self._attributes

    @property
    def node_id(self) -> int:
        return cast(int, self._attributes[NNCFNode.ID_NODE_ATTR])

    @property
    def node_key(self) -> str:
        return cast(str, self._attributes[NNCFNode.KEY_NODE_ATTR])

    @property
    def node_name(self) -> NNCFNodeName:
        return cast(NNCFNodeName, self._attributes[NNCFNode.NODE_NAME_ATTR])

    @property
    def metatype(self) -> Type[OperatorMetatype]:
        return cast(Type[OperatorMetatype], self._attributes[NNCFNode.METATYPE_ATTR])

    @property
    def node_type(self) -> str:
        return cast(str, self._attributes[NNCFNode.NODE_TYPE_ATTR])

    @property
    def layer_name(self) -> Optional[LayerName]:
        return self._attributes.get(NNCFNode.LAYER_NAME_ATTR)

    @layer_name.setter
    def layer_name(self, value: str) -> None:
        self._attributes[NNCFNode.LAYER_NAME_ATTR] = value

    @property
    def layer_attributes(self) -> Optional[BaseLayerAttributes]:
        return self._attributes.get(NNCFNode.LAYER_ATTRIBUTES)

    @layer_attributes.setter
    def layer_attributes(self, value: BaseLayerAttributes) -> None:
        self._attributes[NNCFNode.LAYER_ATTRIBUTES] = value

    @property
    def ignored_algorithms(self) -> List[str]:
        return cast(List[str], self._attributes[NNCFNode.IGNORED_ALGOS_ATTR])

    def is_in_iteration_scope(self) -> bool:
        return cast(bool, self._attributes[NNCFNode.IS_IN_ITERATION_SCOPE_NODE_ATTR])

    def is_integer_input(self) -> bool:
        return cast(bool, self._attributes[NNCFNode.IS_INTEGER_INPUT_NODE_ATTR])

    def is_shared(self) -> bool:
        return cast(bool, self._attributes[NNCFNode.IS_SHARED_ATTR])

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return " ".join([str(self.node_id), self.node_name, self.node_type])

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NNCFNode) and self.attributes == other.attributes


class NNCFGraphEdge:
    """
    A structure describing an edge in NNCFGraph. Since nodes of the NNCFGraph are operations
    on (activation) tensors, an edge in NNCFGraph is a representation of an activation tensor produced or
    consumed by an operation.
    """

    def __init__(
        self,
        from_node: NNCFNode,
        to_node: NNCFNode,
        input_port_id: int,
        output_port_id: int,
        tensor_shape: List[int],
        dtype: Dtype,
        parallel_input_port_ids: List[int],
    ) -> None:
        """
        :param from_node: An NNCFNode that sources the directed edge.
        :param to_node: An NNCFNode that sinks the directed edge.
        :param input_port_id: The ID of the tensor input to the `to_node` that this edge corresponds to.
        :param output_port_id: The ID of the tensor output of the `from_node` that this edge corresponds to..
        :param tensor_shape: The shape of the activation tensor the edge represents.
        :param dtype: The data type of the activation tensor the edge represents.
        """
        self.from_node = from_node
        self.to_node = to_node
        self.input_port_id = input_port_id
        self.output_port_id = output_port_id
        self.tensor_shape: Tuple[int, ...] = tuple(tensor_shape)
        self.dtype = dtype
        self.parallel_input_port_ids = parallel_input_port_ids

    def __str__(self) -> str:
        return f"{self.from_node}:{self.output_port_id} -> {self.tensor_shape} -> {self.to_node}:{self.input_port_id}"

    def __hash__(self) -> int:
        return hash(
            (
                self.from_node,
                self.to_node,
                self.input_port_id,
                self.output_port_id,
                tuple(self.tensor_shape),
                self.dtype,
                tuple(self.parallel_input_port_ids),
            )
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NNCFGraphEdge) and self.__dict__ == other.__dict__


class NNCFGraphPatternIO:
    """
    Describes the inputs and outputs of a subgraph in NNCFGraph.
    """

    def __init__(self, input_edges: List[NNCFGraphEdge], output_edges: List[NNCFGraphEdge]):
        self.input_edges = input_edges
        self.output_edges = output_edges


class NNCFGraph:
    """
    Wrapper over a regular directed acyclic graph that represents a control flow/execution graph of a DNN
    providing some useful methods for graph traversal.
    """

    ACTIVATION_SHAPE_EDGE_ATTR = "activation_shape"
    INPUT_PORT_ID_EDGE_ATTR = "input_port_id"
    OUTPUT_PORT_ID_EDGE_ATTR = "output_port_id"
    DTYPE_EDGE_ATTR = "dtype"
    PARALLEL_INPUT_PORT_IDS_ATTR = "parallel_input_ports"

    def __init__(self) -> None:
        self._nx_graph = nx.DiGraph()
        self._node_id_to_key_dict: Dict[int, str] = {}
        self._nodes: Dict[str, NNCFNode] = {}
        self._input_nncf_nodes: Dict[int, NNCFNode] = {}
        self._output_nncf_nodes: Dict[int, NNCFNode] = {}
        self._node_ids_vs_layer_names: Dict[int, LayerName] = {}
        self._layer_name_vs_shared_nodes: Dict[LayerName, List[NNCFNode]] = defaultdict(list)
        self._node_name_to_node_id_map: Dict[str, List[int]] = {}

    @property
    def nodes(self) -> Dict[str, NNCFNode]:
        return self._nodes

    def get_node_by_id(self, node_id: int) -> NNCFNode:
        """
        :param node_id: Id of the node.
        :return: Node in a graph with such id.
        """
        return self.get_node_by_key(self.get_node_key_by_id(node_id))

    def get_node_by_key(self, key: str) -> NNCFNode:
        """
        :param key: key (node_name) of the node.
        :return: NNCFNode in a graph with such key.
        """
        return self._nodes[key]

    def get_input_nodes(self) -> List[NNCFNode]:
        """
        :return: List of input nodes of the graph.
        """
        return list(self._input_nncf_nodes.values())

    def get_output_nodes(self) -> List[NNCFNode]:
        """
        :return: List of output nodes of the graph.
        """
        return list(self._output_nncf_nodes.values())

    def get_nodes_by_types(self, type_list: List[str]) -> List[NNCFNode]:
        """
        :param type_list: List of types to look for.
        :return: List of nodes with provided types.
        """
        all_nodes_of_type: List[NNCFNode] = []
        if not type_list:
            return all_nodes_of_type
        for nncf_node in self.nodes.values():
            if nncf_node.node_type in type_list:
                all_nodes_of_type.append(nncf_node)
        return all_nodes_of_type

    def get_nodes_by_metatypes(self, metatype_list: Collection[Type[OperatorMetatype]]) -> List[NNCFNode]:
        """
        Return a list of nodes with provided metatypes.

        :param metatype_list: List of types to look for.
        :return: List of nodes with provided metatypes.
        """
        all_nodes_of_type = []
        for nncf_node in self.get_all_nodes():
            if nncf_node.metatype in metatype_list:
                all_nodes_of_type.append(nncf_node)
        return all_nodes_of_type

    def get_all_node_ids(self) -> KeysView[int]:
        """
        Returns all graph nodes' node_ids.
        """
        return self._node_id_to_key_dict.keys()

    def get_all_node_keys(self) -> ValuesView[str]:
        """
        Returns all graph nodes' keys i.e. node_names.
        """
        return self._node_id_to_key_dict.copy().values()

    def get_all_nodes(self) -> List[NNCFNode]:
        """
        Returns list of all graph nodes.
        """
        return list(self._nodes.values())

    def get_all_simple_paths(
        self, start_node_name: NNCFNodeName, end_node_name: NNCFNodeName
    ) -> Generator[List[str], None, None]:
        """
        Generates all simple paths in the NNCFGraph from start node to end node.
        A simple path is a path with no repeated nodes.

        :param start_node_name: a name of starting node for path
        :param end_node_name: a name of node at which to end path
        :return: A generator that produces lists of simple paths. If there are no paths between the start and end nodes
        the generator produces no output.
        """
        start_node = self.get_node_by_name(start_node_name)
        end_node = self.get_node_by_name(end_node_name)
        start_node_key = self.get_node_key_by_id(start_node.node_id)
        end_node_key = self.get_node_key_by_id(end_node.node_id)
        return cast(Generator[List[str], None, None], nx.all_simple_paths(self._nx_graph, start_node_key, end_node_key))

    @staticmethod
    def _get_edge_boundaries(
        match: List[str], graph: nx.DiGraph
    ) -> Tuple[List[Tuple[str, str, Dict[str, Any]]], List[Tuple[str, str, Dict[str, Any]]]]:
        out_edge_boundary = list(nx.edge_boundary(graph, match, data=True))
        complement = list(filter(lambda x: x not in match, graph.nodes.keys()))
        in_edge_boundary = list(nx.edge_boundary(graph, complement, data=True))
        return sorted(in_edge_boundary), sorted(out_edge_boundary)  # must be sorted for determinism

    def get_node_key_by_id(self, node_id: int) -> str:
        """
        Returns node key (node_name) by provided id.

        :param node_id: Id of the node.
        :return: Key of the node with provided id.
        """
        return self._node_id_to_key_dict[node_id]

    def get_next_nodes(self, node: NNCFNode) -> List[NNCFNode]:
        """
        Returns consumer nodes of provided node.

        :param node: Producer node.
        :return: List of consumer nodes of provided node.
        """
        nx_node_keys = self._nx_graph.succ[self._node_id_to_key_dict[node.node_id]]
        return [self._nodes[key] for key in nx_node_keys]

    def get_previous_nodes(self, node: NNCFNode) -> List[NNCFNode]:
        """
        Returns producer nodes of provided node.

        :param node: Consumer node.
        :return: List of producers nodes of provided node.
        """

        nx_node_keys = self._nx_graph.pred[self._node_id_to_key_dict[node.node_id]]
        return [self._nodes[key] for key in nx_node_keys]

    def get_input_edges(self, node: NNCFNode) -> List[NNCFGraphEdge]:
        """
        Returns edges of input tensors with description sorted by input port ID.

        :param node: Consumer node.
        :return: List of input edges for the node sorted by input port ID.
        """
        input_nodes = self.get_previous_nodes(node)
        edges = []
        for from_node in input_nodes:
            edges.extend(self._get_edges(from_node, node))
        return sorted(edges, key=lambda x: x.input_port_id)

    def get_input_edge_by_port_id(self, node: NNCFNode, port_id: int) -> NNCFGraphEdge:
        """
        Returns the input edge for a given node, where edge.input_port_id == port_id is True.

        :param node: The node for which to retrieve the input edge.
        :param port_id: The ID of the input port to filter the edges.
        :return: An input edge connected to the specified input port ID of the
            given node.
        """
        edges = [e for e in self.get_input_edges(node) if e.input_port_id == port_id]
        if len(edges) == 0:
            raise nncf.ValidationError(
                f"Node {node.node_name} does not contain input edge connected to {port_id} port ID."
            )

        if len(edges) > 1:
            raise nncf.InternalError(
                "Unsupported graph. More than one edge was found for a given node by the specified input port ID."
            )
        return edges[0]

    def get_output_edges(self, node: NNCFNode) -> List[NNCFGraphEdge]:
        """
        Returns edges of output tensors sorted by output port ID.

        :param node: Producer node.
        :return: List of output edges for the node sorted by output port ID.
        """
        output_nodes = self.get_next_nodes(node)
        edges = []
        for to_node in output_nodes:
            edges.extend(self._get_edges(node, to_node))
        return sorted(edges, key=lambda x: x.output_port_id)

    def get_output_edges_by_port_id(self, node: NNCFNode, port_id: int) -> List[NNCFGraphEdge]:
        """
        Returns a list of output edges for a given node, filtered by the specified
        output port ID (edge.output_port_id == port_id).

        :param node: The node for which to retrieve the output edges.
        :param port_id: The ID of the output port to filter the edges.
        :return: A list of the output edges connected to the specified output port ID
            of the given node.
        """
        return [e for e in self.get_output_edges(node) if e.output_port_id == port_id]

    def _get_edges(self, from_node: NNCFNode, to_node: NNCFNode) -> List[NNCFGraphEdge]:
        edges = []
        edge = self.get_edge(from_node, to_node)
        parallel_input_port_ids = edge.parallel_input_port_ids
        edge.parallel_input_port_ids = []
        edges.append(edge)
        for input_port_id in parallel_input_port_ids:
            edges.append(
                NNCFGraphEdge(
                    from_node=edge.from_node,
                    to_node=edge.to_node,
                    input_port_id=input_port_id,
                    output_port_id=edge.output_port_id,
                    tensor_shape=list(edge.tensor_shape),
                    dtype=edge.dtype,
                    parallel_input_port_ids=[],
                )
            )
        return edges

    def traverse_graph(
        self,
        curr_node: NNCFNode,
        traverse_function: Callable[[NNCFNode, List[Any]], Tuple[bool, List[Any]]],
        traverse_forward: bool = True,
    ) -> List[Any]:
        """
        Traverses graph up or down starting form `curr_node` node.

        :param curr_node: Node from which traversal is started.
        :param traverse_function: Function describing condition of traversal continuation/termination.
        :param traverse_forward: Flag specifying direction of traversal.
        :return:
        """
        output: List[Any] = []
        return self._traverse_graph_recursive_helper(curr_node, traverse_function, output, traverse_forward)

    def _traverse_graph_recursive_helper(
        self,
        curr_node: NNCFNode,
        traverse_function: Callable[[NNCFNode, List[Any]], Tuple[bool, List[Any]]],
        output: List[Any],
        traverse_forward: bool,
    ) -> List[Any]:
        is_finished, output = traverse_function(curr_node, output)
        get_nodes_fn = self.get_next_nodes if traverse_forward else self.get_previous_nodes
        if not is_finished:
            for node in get_nodes_fn(curr_node):
                self._traverse_graph_recursive_helper(node, traverse_function, output, traverse_forward)
        return output

    def add_nncf_node(
        self,
        node_name: str,
        node_type: str,
        node_metatype: Type[OperatorMetatype],
        layer_attributes: Optional[BaseLayerAttributes] = None,
        node_id_override: Optional[int] = None,
        layer_name: Optional[LayerName] = None,
        ignored_algorithms: Optional[List[str]] = None,
        is_in_iteration_scope: bool = False,
        is_integer_input: bool = False,
        is_shared: bool = False,
    ) -> NNCFNode:
        """
        Adds a node into the graph. A node represents an operation being performed on tensors.
        :param node_name: The name of the node to add - will serve as a human-readable and specifiable ID.
        :param node_type: The type of the node - usually a framework-specific string representation of the operation.
        :param node_metatype: The metatype of the node - a framework-abstract definition of what the operation
            actually means.
        :param layer_attributes: Must be passed for operations that, in order to be processed
            during compression, require additional information such as the exact dimension of the weights tensor to be
            considered a "channel" dimension for per-channel quantization, the weight shape itself for sparsity mask
            creation etc.
        :param node_id_override: The numerical ID to be associated with the new node; if unspecified, will
            assign a unique ID.
        :param layer_name: The name of the framework-specific "layer" object that houses the operation represented by
            the node and associated trainable weights, if any.
        :param ignored_algorithms: A list of compression algorithm names (from the same set of strings that are
            specified in the `"algorithm": ...` section of the .json NNCF config or `ptq_quantization`)
            which should ignore this operation.
        :param is_in_iteration_scope: Whether the node to be currently added corresponds to an iteration of an RNN
            cycle (where the number of iterations is determined dynamically based on the RNN input shape).
        :param is_integer_input: Only valid for input nodes - whether the input node corresponds to an integer input.
        :param is_shared: Whether the node corresponds to an operation that accesses the weights that are also accessed
            by another operation (represented by a separate node) in NNCFGraph. Examples would be repeated applications
            of a convolution layer with the same weights at different stages in the network.
        :return: An NNCFNode object representing the newly created node in the graph. The node will still have
            to be connected to the rest of the nodes with edges using the `NNCFGraph.add_edge_between_nncf_nodes`
            method.
        """
        if node_id_override is not None:
            node_id = node_id_override
        else:
            node_ids = list(self.get_all_node_ids())
            if node_ids:
                node_id = max(self.get_all_node_ids()) + 1
            else:
                node_id = 0

        if node_id in self._node_id_to_key_dict:
            raise ValueError(f"NNCF node with id {node_id} is already in the NNCFGraph")

        node_ids = self._node_name_to_node_id_map.setdefault(node_name, [])
        node_ids.append(node_id)

        node_key = f"{node_id} {node_name}"

        self._node_id_to_key_dict[node_id] = node_key
        attrs = {
            NNCFNode.ID_NODE_ATTR: node_id,
            NNCFNode.NODE_NAME_ATTR: node_name,
            NNCFNode.KEY_NODE_ATTR: node_key,
            NNCFNode.NODE_TYPE_ATTR: node_type,
            NNCFNode.LAYER_NAME_ATTR: layer_name,
            NNCFNode.METATYPE_ATTR: node_metatype,
            NNCFNode.IS_SHARED_ATTR: is_shared,
            NNCFNode.IS_IN_ITERATION_SCOPE_NODE_ATTR: is_in_iteration_scope,
            NNCFNode.IS_INTEGER_INPUT_NODE_ATTR: is_integer_input,
        }
        if layer_attributes is not None:
            attrs[NNCFNode.LAYER_ATTRIBUTES] = layer_attributes

        if ignored_algorithms is None:
            ignored_algorithms = []
        attrs[NNCFNode.IGNORED_ALGOS_ATTR] = ignored_algorithms
        self._nx_graph.add_node(node_key, **attrs)

        node = NNCFNode(self._nx_graph.nodes[node_key])
        self._nodes[node_key] = node

        if node.metatype in INPUT_NOOP_METATYPES:
            self._input_nncf_nodes[node_id] = node

        if node.metatype in OUTPUT_NOOP_METATYPES:
            self._output_nncf_nodes[node_id] = node

        if layer_name is not None:
            self._node_ids_vs_layer_names[node.node_id] = layer_name
            self._layer_name_vs_shared_nodes[layer_name].append(node)

        return node

    def add_edge_between_nncf_nodes(
        self,
        from_node_id: int,
        to_node_id: int,
        tensor_shape: Union[Tuple[int, ...], List[int]],
        input_port_id: int,
        output_port_id: int,
        dtype: Dtype,
        parallel_input_port_ids: Optional[List[int]] = None,
    ) -> None:
        """
        Adds a directed edge between two `NNCFNode`s that are already present in the graph.
        The edge represents an activation tensor, produced or consumed by an operation (which is represented by a node)
        :param from_node_id: The `NNCFNode.node_id` of the node that produces the tensor represented by the edge.
        :param to_node_id: The `NNCFNode.node_id` of the node that consumes the tensor represented by the edge.
        :param tensor_shape: The shape of the tensor represented by the edge.
        :param input_port_id: Specifies the index among the possible inputs of the `to_node_id` node' that this tensor
            should correspond to.
        :param output_port_id: Specifies the index among the possible outputs of the `from_node_id` node' that this
            tensor should correspond to.
        :param dtype: The data type of the tensor.
        :param parallel_input_port_ids: Input ports for parallel edges, if any should be present for this edge.
        """
        from_node_key = self._node_id_to_key_dict[from_node_id]
        to_node_key = self._node_id_to_key_dict[to_node_id]

        err_reason = None

        if from_node_key not in self._nx_graph.nodes:
            err_reason = f"node {from_node_key} not in NNCFGraph"
        if to_node_key not in self._nx_graph.nodes:
            err_reason = f"node {from_node_key} not in NNCFGraph"
        if from_node_id in self._output_nncf_nodes:
            err_reason = "cannot add edges *from* output nodes"
        if to_node_id in self._input_nncf_nodes:
            err_reason = "cannot add edges *to* input nodes"

        if err_reason is not None:
            raise ValueError(f"Cannot add edge from {from_node_key} to {to_node_key} - {err_reason}!")

        attrs = {
            NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR: tensor_shape,
            NNCFGraph.INPUT_PORT_ID_EDGE_ATTR: input_port_id,
            NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR: output_port_id,
            NNCFGraph.DTYPE_EDGE_ATTR: dtype,
            NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR: [] if parallel_input_port_ids is None else parallel_input_port_ids,
        }
        self._nx_graph.add_edge(from_node_key, to_node_key, **attrs)

    def topological_sort(self) -> List[NNCFNode]:
        """
        Returns nodes in topologically sorted order, additionally sorted in ascending node ID order.
        """
        return [
            self._nodes[node_name]
            for node_name in nx.lexicographical_topological_sort(
                self._nx_graph, key=lambda x: self._nx_graph.nodes[x][NNCFNode.ID_NODE_ATTR]
            )
        ]

    def dump_graph(self, path: str) -> None:
        write_dot_graph(self.get_graph_for_structure_analysis(), pathlib.Path(path))

    def visualize_graph(self, path: str) -> None:
        out_graph = self._get_graph_for_visualization()
        write_dot_graph(out_graph, pathlib.Path(path))

    def get_graph_for_structure_analysis(self, extended: bool = False) -> nx.DiGraph:
        """
        Returns the nx.Digraph, which is built based on self._nx_graph.
        The new graph has certain node attributes omitted, compared to the graph stored inside NNCFGraph.
        If the node name consists of a special reserved character, this character will be replaced.

        :param extended: whether the graph edges should have attributes: shape of the tensor and tensor primitive type.
        :return: An nx.DiGraph to be used for structure analysis
        """
        out_graph = nx.DiGraph()
        for node_name, node in self._nx_graph.nodes.items():
            attrs_node = {"id": str(node[NNCFNode.ID_NODE_ATTR]), "type": node[NNCFNode.NODE_TYPE_ATTR]}
            for attr in ["color", "label", "style"]:
                if attr in node:
                    attrs_node[attr] = node[attr]

            out_graph.add_node(node_name, **attrs_node)

        for u, v in self._nx_graph.edges:
            edge = self._nx_graph.edges[u, v]
            attrs_edge = {}
            label = {}
            if edge[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR]:
                label["parallel_input_port_ids"] = edge[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR]

            if extended:
                if edge[NNCFGraph.DTYPE_EDGE_ATTR] is Dtype.INTEGER:
                    attrs_edge["style"] = "dashed"
                else:
                    attrs_edge["style"] = "solid"
                label["shape"] = edge[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]

            if label:
                if "shape" in label and len(label) == 1:
                    attrs_edge["label"] = label["shape"]
                else:
                    attrs_edge["label"] = ", ".join((f"{k} {v}" for k, v in label.items()))
            out_graph.add_edge(u, v, **attrs_edge)
        return out_graph

    def _get_graph_for_visualization(self) -> nx.DiGraph:
        """
        :return: A user-friendly graph .dot file, making it easier to debug the network and setup
        ignored/target scopes.
        """
        out_graph = nx.DiGraph()
        for node in self.get_all_nodes():
            attrs_node = {}
            attrs_node["label"] = f"{node.node_id} {node.node_name}"
            node_key = self.get_node_key_by_id(node.node_id)
            out_graph.add_node(node_key, **attrs_node)

        for u, v in self._nx_graph.edges:
            edge = self._nx_graph.edges[u, v]
            if edge[NNCFGraph.DTYPE_EDGE_ATTR] is Dtype.INTEGER:
                style = "dashed"
            else:
                style = "solid"
            edge_label = (
                f"{edge[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]} \\n"
                f"{edge[NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR]} -> {edge[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR]}"
            )
            out_graph.add_edge(u, v, label=edge_label, style=style)

        mapping = {k: v["label"] for k, v in out_graph.nodes.items()}
        out_graph = nx.relabel_nodes(out_graph, mapping)
        for node in out_graph.nodes.values():
            node.pop("label")  # type: ignore

        return out_graph

    def get_node_by_name(self, name: NNCFNodeName) -> NNCFNode:
        node_ids = self._node_name_to_node_id_map.get(name, None)
        if node_ids is None:
            raise nncf.InternalError("Could not find a node {} in NNCFGraph!".format(name))
        if len(node_ids) > 1:
            raise nncf.InternalError(f"More than one node in NNCFGraph matches name {name}")

        node_key = f"{node_ids[0]} {name}"
        return self._nodes[node_key]

    def __eq__(self, other: object) -> bool:
        nm = iso.categorical_node_match(
            [NNCFNode.ID_NODE_ATTR, NNCFNode.KEY_NODE_ATTR, NNCFNode.LAYER_ATTRIBUTES], [None, None, None]
        )
        em = iso.categorical_edge_match(
            [NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR, NNCFGraph.INPUT_PORT_ID_EDGE_ATTR], [None, None]
        )
        return isinstance(other, NNCFGraph) and bool(
            nx.is_isomorphic(self._nx_graph, other._nx_graph, node_match=nm, edge_match=em)
        )

    def get_nx_graph_copy(self) -> nx.DiGraph:
        return deepcopy(self._nx_graph)

    def get_nncf_graph_pattern_io(self, match: List[str]) -> NNCFGraphPatternIO:
        """
        Returns an NNCFGraphPatternIO object that describes the input/output nodes and edges of a
        subgraph specified by `match`.

        :param match: A list of node keys specifying a subgraph to be matched. The subgraph to be matched will
        consist of nodes with the same keys that are connected with edges in the order they are listed in the
        `match` list
        :return: NNCFGraphPatternIO object describing the inputs and outputs of the matched subgraph
        """

        in_edge_boundary, out_edge_boundary = NNCFGraph._get_edge_boundaries(match, self._nx_graph)
        boundary = in_edge_boundary + out_edge_boundary
        input_nncf_edges = []
        output_nncf_edges = []

        for nx_edge in boundary:
            from_node_key = nx_edge[0]
            to_node_key = nx_edge[1]
            data = nx_edge[2]
            nncf_edge = NNCFGraphEdge(
                self._nodes[from_node_key],
                self._nodes[to_node_key],
                input_port_id=data[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR],
                output_port_id=data[NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR],
                tensor_shape=data[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR],
                dtype=data[NNCFGraph.DTYPE_EDGE_ATTR],
                parallel_input_port_ids=data[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR],
            )
            if from_node_key in match:
                output_nncf_edges.append(nncf_edge)
            elif to_node_key in match:
                input_nncf_edges.append(nncf_edge)
            else:
                raise nncf.InternalError("Invalid graph expression supplied!")

        return NNCFGraphPatternIO(input_nncf_edges, output_nncf_edges)

    def get_nx_edge(self, node_u: NNCFNode, node_v: NNCFNode) -> OutEdgeView:
        nx_node_u = self._nx_graph.nodes[self._node_id_to_key_dict[node_u.node_id]]
        nx_node_v = self._nx_graph.nodes[self._node_id_to_key_dict[node_v.node_id]]
        return self._nx_graph.edges[nx_node_u["key"], nx_node_v["key"]]

    def get_nodes_count(self) -> int:
        return int(self._nx_graph.number_of_nodes())

    def get_edge(self, from_node: NNCFNode, to_node: NNCFNode) -> NNCFGraphEdge:
        """
        Returns an NNCFGraphEdge object that corresponds to an edge connecting two given NNCFNodes in this
        graph.
        :param from_node: The NNCFNode in this graph that sources the edge.
        :param to_node: The NNCFNode in this graph that sinks the edge.
        :return: The NNCFGraphEdge object representing the edge between `from_node` and `to_node`.
        """
        data = self.get_nx_edge(from_node, to_node)
        return NNCFGraphEdge(
            from_node,
            to_node,
            data[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR],
            data[NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR],
            data[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR],
            data[NNCFGraph.DTYPE_EDGE_ATTR],
            data[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR],
        )

    def get_all_edges(self) -> Generator[NNCFGraphEdge, None, None]:
        for nx_edge in self._nx_graph.in_edges:
            yield self.get_edge(self.get_node_by_key(nx_edge[0]), self.get_node_by_key(nx_edge[1]))

    def remove_nodes_from(self, nodes: Collection[NNCFNode]) -> None:
        """
        Removes nodes from the current NNCFGraph instance.
        We use the remove_node method here because remove_nodes_from uses a silent fail instead of an exception.

        :param nodes: List of NNCFNodes to remove.
        """
        for node in nodes:
            self._nx_graph.remove_node(node.node_key)
            del self._nodes[node.node_key]

        self._node_id_to_key_dict = {}
        for node_key, node in self._nx_graph.nodes.items():
            self._node_id_to_key_dict[node["id"]] = node_key  # type:ignore

    def find_matching_subgraphs(self, patterns: GraphPattern, strict: bool = True) -> List[List[NNCFNode]]:
        """
        Returns subgraphs of matched pattern in patterns.

        :param patterns: Instance of GraphPattern containing all patterns.
        :param strict: If True returns only strict matched subgraphs, if False - all matched subgraphs.
        :return: List of subgraphs that are matching by pattern matching.
            Subgraph is a ordered list of nodes of matched subgraph
        The returned nodes order relies on DiGraphMatcher isomorphic subgraphs matching logic from networkX package.
        DiGraphMatcher does not guarantee a specific order for returning isomorphic subgraphs.
        """
        output = []
        for matched_subgraph in find_subgraphs_matching_pattern(self._nx_graph, patterns, strict):
            subgraph_list = []
            for node_key in matched_subgraph:
                subgraph_list.append(self.get_node_by_key(node_key))
            output.append(subgraph_list)
        return output



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TypeVar

from nncf.experimental.quantization.algorithms.range_estimator.algorithm import MinMaxRangeEstimator
from nncf.experimental.quantization.quantizer import Quantizer
from nncf.quantization.advanced_parameters import AdvancedBiasCorrectionParameters
from nncf.quantization.advanced_parameters import AdvancedSmoothQuantParameters
from nncf.quantization.advanced_parameters import RangeEstimatorParameters
from nncf.quantization.algorithms.bias_correction.algorithm import BIAS_CORRECTION_THRESHOLD
from nncf.quantization.algorithms.bias_correction.algorithm import BiasCorrection
from nncf.quantization.algorithms.fast_bias_correction.algorithm import FAST_BIAS_CORRECTION_THRESHOLD
from nncf.quantization.algorithms.fast_bias_correction.algorithm import FastBiasCorrection
from nncf.quantization.algorithms.pipeline import Pipeline
from nncf.quantization.algorithms.smooth_quant.algorithm import SmoothQuant

TModel = TypeVar("TModel")


def experimental_create_ptq_pipeline(
    quantizer: Quantizer,
    subset_size: int = 300,
    fast_bias_correction: Optional[bool] = True,
    smooth_quant: bool = False,
    bias_correction_params: Optional[AdvancedBiasCorrectionParameters] = None,
    smooth_quant_params: Optional[AdvancedSmoothQuantParameters] = None,
    activations_range_estimator_params: Optional[RangeEstimatorParameters] = None,
    weights_range_estimator_params: Optional[RangeEstimatorParameters] = None,
    batchwise_statistics: bool = False,
) -> Pipeline:
    """
    Creates an experimental post-training quantization pipeline.

    The experimental post-training quantization pipeline includes the following steps:
        1) SmoothQuant
        2) MinMaxRangeInit
        3) FastBiasCorrection or BiasCorrection

    :param quantizer: Quantizer to use in MiMaxRangeInit algorithm.
    :param subset_size: Size of a subset to calculate activations
        statistics used for quantization.
    :param fast_bias_correction: Setting this option to `False` enables a different
        bias correction method which is more accurate, in general, and takes
        more time but requires less memory. None disables the bias correction algorithm.
    :param smooth_quant: Setting this option to `True` enables the SmoothQuant algorithm.
    :param bias_correction_params: Contains advanced parameters for fine-tuning bias correction algorithm.
    :param smooth_quant_params: Contains advanced alpha parameters for SmoothQuant algorithm.
    :param activations_range_estimator_params: Contains parameters for estimating the range
        of activations of the model.
    :param weights_range_estimator_params: Contains parameters for estimating the range
        of weights of the model.
    :param batchwise_statistics: Determines whether quantizer statistics should be calculated
        for each item of the batch or for the entire batch, default is False.
    :return: An experimental post-training quantization pipeline.
    """

    # Build the post-training quantization pipeline.
    pipeline_steps = []

    if smooth_quant_params is None:
        smooth_quant_params = AdvancedSmoothQuantParameters()

    if smooth_quant and (smooth_quant_params.convolution >= 0 or smooth_quant_params.matmul >= 0):
        alpha_map = {"convolution": smooth_quant_params.convolution, "matmul": smooth_quant_params.matmul}
        pipeline_steps.append([SmoothQuant(subset_size, False, alpha_map=alpha_map)])

    # Add the `MinMaxQuantization` algorithm as the third step of the pipeline.
    pipeline_steps.append(
        [
            MinMaxRangeEstimator(
                quantizer=quantizer,
                subset_size=subset_size,
                inplace_statistics=False,
                batchwise_statistics=batchwise_statistics,
                activations_range_estimator_params=activations_range_estimator_params,
                weights_range_estimator_params=weights_range_estimator_params,
            )
        ]
    )

    if fast_bias_correction is not None:
        # Add the `FastBiasCorrection` or `BiasCorrection` as additional algorithm
        # inside the third step of the pipeline. It is added after `MinMaxQuantization`
        # algorithm.
        if fast_bias_correction:
            threshold = FAST_BIAS_CORRECTION_THRESHOLD
            bias_correction_subset_size = subset_size
            bias_correction_cls = FastBiasCorrection
        else:
            threshold = BIAS_CORRECTION_THRESHOLD
            bias_correction_subset_size = max(int(subset_size * 0.2), 1)
            bias_correction_cls = BiasCorrection

        if bias_correction_params is None:
            bias_correction_params = AdvancedBiasCorrectionParameters()

        if bias_correction_params.threshold is not None:
            threshold = bias_correction_params.threshold

        pipeline_steps[-1].append(
            bias_correction_cls(
                bias_correction_subset_size,
                threshold,
                bias_correction_params.apply_for_all_nodes,
            )
        )

    return Pipeline(pipeline_steps)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Union

import torch.fx

from nncf.common.graph.transformations.commands import Command
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.transformations.commands import TransformationType


class FXApplyTransformationCommand(Command):
    """
    Command to apply given transformation to a model.
    """

    def __init__(
        self,
        transformation_fn: Callable[[torch.fx.GraphModule], None],
        priority: Union[TransformationPriority, int] = TransformationPriority.DEFAULT_PRIORITY,
    ):
        """
        :param transformation_fn: Target transformation function.
        :param priority: Transformation priority.
        """
        super().__init__(TransformationType.INSERT)
        self.transformation_fn = transformation_fn
        self.priority = priority



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Optional, Tuple, Union

import torch

from nncf.common.graph.layer_attributes import Dtype


class TensorMeta:
    @staticmethod
    def default_comparator(lhs: "TensorMeta", rhs: "TensorMeta"):
        return lhs.index == rhs.index and lhs.creator_id == rhs.creator_id and lhs.shape[1:] == rhs.shape[1:]

    def __init__(
        self,
        creator_id: Union[int, None],
        index: int,
        shape: Union[List[int], Tuple[torch.Tensor, ...]],
        dtype: Dtype = Dtype.FLOAT,
    ):
        """
        :param creator_id: An ID of the node in DynamicGraph that corresponds to an operation that created the
            tensor.
        :param index: The index of this tensor in the creator operation's output.
        :param shape: The shape of the tensor.
        :param dtype: The type of the tensor
        """
        self.creator_id = creator_id
        self.index = index
        self.shape = tuple(int(dim) for dim in shape)  # Handle cases when shape is a tuple of Tensors
        self.dtype = dtype

    def __eq__(self, other):
        if not isinstance(other, TensorMeta):
            return False
        return self.default_comparator(self, other)

    def __hash__(self):
        return hash((self.creator_id, self.index, self.shape))

    def __str__(self):
        return "C{}_I{}_".format(self.creator_id, self.index) + "S" + "x".join([str(s) for s in self.shape])


class TracedTensorMixin:
    """
    A mixin class providing tracing capabilities to PyTorch tensors.

    This class provides interfaces for patching a given torch tensor and associating it with the provided tensor_meta.
    """

    TENSOR_META = "tensor_meta"
    ORIGINAL_CLASS = "original_class"
    _TRACING_ATTRS = "_tracing_attrs"

    @property
    def tensor_meta(self):
        return self.tracing_attrs[TracedTensorMixin.TENSOR_META]

    @tensor_meta.setter
    def tensor_meta(self, value: Union[None, TensorMeta]):
        self.tracing_attrs[TracedTensorMixin.TENSOR_META] = value

    @property
    def tracing_attrs(self):
        if not hasattr(self, TracedTensorMixin._TRACING_ATTRS):
            self._tracing_attrs = {}
        return self._tracing_attrs

    @classmethod
    def patch(cls, tensor: torch.Tensor, tensor_meta: Optional[TensorMeta] = None) -> "TracedTensorMixin":
        """
        Patches a tensor with the TracedTensorMixin interface and associates it with the provided tensor_meta.

        :param tensor: The input tensor.
        :param tensor_meta: The metadata associated with the tensor.
        :return: The patched ternsor.
        """
        if not isinstance(tensor, TracedTensorMixin):
            original_class = tensor.__class__
            tensor.__class__ = cls
            tensor.tracing_attrs[TracedTensorMixin.ORIGINAL_CLASS] = original_class

        tensor.tensor_meta = tensor_meta
        return tensor

    def strip(self) -> None:
        """
        Reverts the tensor to its original class by removing tracing attributes.
        """
        self.__class__ = self.tracing_attrs[TracedTensorMixin.ORIGINAL_CLASS]
        delattr(self, TracedTensorMixin._TRACING_ATTRS)


class TracedTensor(torch.Tensor, TracedTensorMixin):
    """
    When tracing a torch model, intermediate tensors will be dynamically turned into
    instances of this class to be able to store additional data required for establishing
    relation between tensor producer and consumer operations.
    """

    @staticmethod
    def from_torch_tensor(tensor: torch.Tensor, tensor_meta: Optional[TensorMeta] = None) -> "TracedTensor":
        """
        Creates a TracedTensor by patching a given torch.Tensor, associating it with the provided tensor_meta.

        :param tensor: The input torch.Tensor.
        :param tensor_meta: The metadata associated with the tensor.
        :return: The resulting TracedTensor.
        """
        return TracedTensor.patch(tensor, tensor_meta)

    def as_subclass(self, cls: "TracedTensor") -> "TracedTensor":
        """
        Required for PyTorch 1.7.0 compatibility - the handle_torch_function and __torch_function__
        API in general calls this after a wrapped function call; need to preserve the tensor_meta extensions
        """

        return self

    # NOTE: This disables the __torch_function__ API altogether when using NNCF.
    # TODO: make NNCF utilize the __torch_function__ API instead.

    if hasattr(torch._C, "_disabled_torch_function_impl"):
        __torch_function__ = torch._C._disabled_torch_function_impl


class TracedParameter(torch.nn.Parameter, TracedTensorMixin):
    """
    When tracing a torch model, model parameters will be dynamically turned into
    instances of this class to be able to store additional data required for tracing parameters
    across operations.
    """

    NAME = "name"
    IS_REUSED = "is_reused"

    @property
    def name(self):
        return self.tracing_attrs[TracedParameter.NAME]

    @property
    def is_reused(self):
        return self.tracing_attrs[TracedParameter.IS_REUSED]

    def get_dtype(self):
        # Type of self is TracedParameter or TracedTensor
        return super(self.__class__, self).__getattribute__("dtype")

    def __getattribute__(self, name):
        if name == "dtype":
            return self.get_dtype()
        return super().__getattribute__(name)

    @staticmethod
    def from_torch_parameter(tensor: torch.nn.Parameter, name: str, is_reused: bool) -> "TracedParameter":
        """
        Creates a TracedParameter by patching a given torch.nn.Parameter, associating it
        with the provided parameter name.

        :param tensor: The input torch.nn.Parameter.
        :param name: The parameter name.
        :param is_reused: True if parameter is used as an input in several operations of the model otherwise False.
        :return: The resulting TracedParameter.
        """
        TracedParameter.patch(tensor)
        tensor.tracing_attrs[TracedParameter.NAME] = name
        tensor.tracing_attrs[TracedParameter.IS_REUSED] = is_reused
        return tensor

    def as_subclass(self, cls: "TracedParameter") -> "TracedParameter":
        """
        Required for PyTorch 1.7.0 compatibility - the handle_torch_function and __torch_function__
        API in general calls this after a wrapped function call; need to preserve the tensor_meta extensions
        """

        return self

    # NOTE: This disables the __torch_function__ API altogether when using NNCF.
    # TODO: make NNCF utilize the __torch_function__ API instead.

    if hasattr(torch._C, "_disabled_torch_function_impl"):
        __torch_function__ = torch._C._disabled_torch_function_impl



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List, Optional, Tuple

from nncf.common.logging import nncf_logger
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.experimental.torch.nas.bootstrapNAS.training.base_training import BNASTrainingAlgorithm
from nncf.experimental.torch.nas.bootstrapNAS.training.lr_scheduler import BaseLRScheduler
from nncf.experimental.torch.nas.bootstrapNAS.training.stage_descriptor import DEFAULT_STAGE_LR_RATE
from nncf.experimental.torch.nas.bootstrapNAS.training.stage_descriptor import StageDescriptor


class NSParamsStateNames:
    LIST_STAGE_DESCRIPTIONS = "list_stage_descriptions"


class NASSchedulerParams:
    _state_names = NSParamsStateNames

    def __init__(self, list_stage_descriptions: Optional[List[StageDescriptor]] = None):
        """
        Constructor

        :param list_stage_descriptions: List of parameters per each supernet training stage.
        """
        if list_stage_descriptions is None:
            list_stage_descriptions = [
                StageDescriptor(train_dims=[ElasticityDim.KERNEL], epochs=1),
                StageDescriptor(train_dims=[ElasticityDim.KERNEL, ElasticityDim.DEPTH], epochs=1),
                StageDescriptor(train_dims=[ElasticityDim.KERNEL, ElasticityDim.DEPTH], epochs=1, depth_indicator=2),
                StageDescriptor(train_dims=[ElasticityDim.KERNEL, ElasticityDim.DEPTH, ElasticityDim.WIDTH], epochs=1),
                StageDescriptor(
                    train_dims=[ElasticityDim.KERNEL, ElasticityDim.DEPTH, ElasticityDim.WIDTH],
                    epochs=1,
                    width_indicator=2,
                    reorg_weights=True,
                    bn_adapt=True,
                ),
            ]
        self.list_stage_descriptions = list_stage_descriptions

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "NASSchedulerParams":
        """
        Creates the object from its config.
        """
        descs_config = config.get(cls._state_names.LIST_STAGE_DESCRIPTIONS)
        descs = None
        if descs_config is not None:
            descs = [StageDescriptor.from_config(stage_desc_config) for stage_desc_config in descs_config]
        return cls(descs)

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "NASSchedulerParams":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        list_stage_descriptions_state = state[cls._state_names.LIST_STAGE_DESCRIPTIONS]
        list_stage_descriptions = [StageDescriptor.from_state(state) for state in list_stage_descriptions_state]
        return cls(list_stage_descriptions)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression loss state.

        :return: The compression loss state.
        """
        return {
            self._state_names.LIST_STAGE_DESCRIPTIONS: [desc.get_state() for desc in self.list_stage_descriptions],
        }

    def __eq__(self, other: "NASSchedulerParams") -> bool:
        return self.__dict__ == other.__dict__


class BNASSchedulerStateNames:
    LIST_STAGE_DESCRIPTIONS = "list_stage_descriptions"


class BootstrapNASScheduler(BaseCompressionScheduler):
    """
    The cornerstone of supernet training within a NAS algorithm. The `step()` and `epoch_step()` methods of the
    compression scheduler must be called in the beginning of each training step and epoch, respectively.
    These methods trigger a subnet activations, elasticity configuration during the training.
    """

    _state_names = BNASSchedulerStateNames

    def __init__(
        self,
        training_ctrl: BNASTrainingAlgorithm,
        params: NASSchedulerParams,
        available_elasticity_dims: List[ElasticityDim],
        progressivity_of_elasticity: List[ElasticityDim],
    ):
        super().__init__()
        self._training_ctrl = training_ctrl
        self._params = params
        self._available_elasticity_dims = available_elasticity_dims
        self._progressivity_of_elasticity = progressivity_of_elasticity
        self.current_stage_idx = -1
        # Property setter with validation is not used intentionally for the resume case. When the actual list stage
        #  descriptors are loaded after creation of the scheduler. Scheduler is resumed without config = with empty
        #  params = default stage descriptors, that could lead to inconsistency with progressivity and enabled dims.
        #  The validation will happen in the first usage of list_stage_descriptors property.
        self._list_stage_descriptors = self._params.list_stage_descriptions
        self._is_elasticity_dims_validated = False
        self._lr_scheduler = None

    @property
    def lr_scheduler(self):
        return self._lr_scheduler

    @lr_scheduler.setter
    def lr_scheduler(self, lr_scheduler: BaseLRScheduler) -> None:
        self._lr_scheduler = lr_scheduler

    @property
    def current_step(self) -> int:
        return self._lr_scheduler.current_step

    @property
    def list_stage_descriptors(self) -> List[StageDescriptor]:
        """
        :return: a list of stage descriptors (parameters of the training stage).
        """
        if not self._is_elasticity_dims_validated:
            self._validate_elasticity_dims(self._available_elasticity_dims, self._progressivity_of_elasticity)
        self._is_elasticity_dims_validated = True
        self._validate_lr()
        return self._list_stage_descriptors

    @list_stage_descriptors.setter
    def list_stage_descriptors(self, stage_descriptors: List[StageDescriptor]) -> None:
        """
        Sets a given stage descriptors to the schedule. Can be used on loading state from a checkpoint.

        :param stage_descriptors: list of stage descriptors
        """
        self._list_stage_descriptors = stage_descriptors
        self._validate_elasticity_dims(self._available_elasticity_dims, self._progressivity_of_elasticity)
        self._is_elasticity_dims_validated = True
        self._validate_lr()

    def step(self, next_step: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training step to prepare
        the compression method to continue training the model in the `next_step`.

        :param next_step: The global step index for which the compression scheduler
            will update the state of the compression method.
        """
        self._training_ctrl.step()
        self._lr_scheduler.step(next_step)

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training epoch to prepare
        the compression method to continue training the model in the `next_epoch`.

        :param next_epoch: The epoch index for which the compression scheduler
            will update the state of the compression method.
        """
        super().epoch_step(next_epoch)
        self._lr_scheduler.epoch_step(next_epoch)
        stage_desc, stage_desc_idx = self.get_current_stage_desc()
        if stage_desc is not None and stage_desc_idx != self.current_stage_idx:
            self._lr_scheduler.stage_step(stage_desc)
            self._training_ctrl.set_stage(stage_desc)
            self.current_stage_idx = stage_desc_idx

    def is_final_stage(self) -> bool:
        """
        :return: True, if final stage has been reached, False - otherwise
        """
        return self.current_stage_idx == len(self.list_stage_descriptors) - 1

    def get_current_stage_desc(self) -> Tuple[Optional[StageDescriptor], int]:
        """
        :return: current stage descriptor and its index in the list of all descriptors
        """
        partial_epochs = 0
        for stage_desc_idx, stage_desc in enumerate(self.list_stage_descriptors):
            partial_epochs += stage_desc.epochs
            if self.current_epoch < partial_epochs:
                return stage_desc, stage_desc_idx
        return None, -1

    def get_total_training_epochs(self) -> int:
        """
        Returns total number of epochs required for the supernet training.

        :return: number of epochs
        """
        total_epochs = 0
        for stage_desc in self.list_stage_descriptors:
            total_epochs += stage_desc.epochs
        return total_epochs

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Loads the compression scheduler state, but does not update the state of the
        compression method.

        :param state: Output of `get_state()` method.
        """
        super().load_state(state)
        list_stage_descriptors = state[self._state_names.LIST_STAGE_DESCRIPTIONS]
        # No conflict resolving with the related config options, parameters are overridden by compression state
        self.list_stage_descriptors = list(map(StageDescriptor.from_state, list_stage_descriptors))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression scheduler state.

        :return: The compression scheduler state.
        """
        state = super().get_state()
        state[self._state_names.LIST_STAGE_DESCRIPTIONS] = [desc.get_state() for desc in self.list_stage_descriptors]
        return state

    def _validate_elasticity_dims(
        self, available_elasticity_dims: List[ElasticityDim], progressivity_of_elasticity: List[ElasticityDim]
    ) -> None:
        last_stage = -1
        first_stage = len(progressivity_of_elasticity)
        for desc in self._list_stage_descriptors:
            high_priority_dim_idx = -1
            low_priority_dim_idx = len(progressivity_of_elasticity)
            stages_covered = []
            for train_dim in desc.train_dims:
                if train_dim not in available_elasticity_dims:
                    raise ValueError(
                        f"Invalid training elasticity dimension {train_dim} in the scheduler.\n"
                        f"The elasticity for this dimension is not enabled.\n"
                        f"It can be enabled by specifying `available_elasticity_dims` param in the `elasticity` "
                        f"section of config.\n"
                        f"List of currently available dimensions: {[dim.value for dim in available_elasticity_dims]}"
                    )
                dim_idx = progressivity_of_elasticity.index(train_dim)
                if dim_idx not in stages_covered:
                    stages_covered.append(dim_idx)
                if dim_idx > high_priority_dim_idx:
                    high_priority_dim_idx = dim_idx
                if dim_idx < low_priority_dim_idx:
                    low_priority_dim_idx = dim_idx
            if high_priority_dim_idx < last_stage or low_priority_dim_idx > first_stage:
                raise ValueError(
                    f"stage {progressivity_of_elasticity[high_priority_dim_idx]} violates progressivity of elasticity"
                )
            for i in range(low_priority_dim_idx, high_priority_dim_idx):
                if i not in stages_covered and progressivity_of_elasticity[i] in available_elasticity_dims:
                    raise ValueError(
                        f"Missed to call {progressivity_of_elasticity[i]} in {desc.train_dims} which violates "
                        f"progressivity of elasticity {progressivity_of_elasticity}"
                    )
            last_stage = high_priority_dim_idx
            first_stage = low_priority_dim_idx

    def _validate_lr(self):
        for desc in self._list_stage_descriptors:
            # Check if global learning rate has been set
            if desc.init_lr is not None and bool(self._training_ctrl.lr_schedule_config):
                raise ValueError(
                    f"Global learning rate scheduler is in use. Cannot set stage learning rate: {desc.init_lr}"
                )
            # Check if stage learning rate has been set
            if desc.init_lr is None and not bool(self._training_ctrl.lr_schedule_config):
                nncf_logger.warning(
                    "Stage learning rate in use but init_lr value for stage wasn't set. Using default value of 3.5e-6"
                )
                desc.init_lr = DEFAULT_STAGE_LR_RATE

            if desc.init_lr is not None and desc.epochs_lr is None:
                nncf_logger.warning(
                    f"Stage learning rate in use but epochs_lr value for stage wasn't set. "
                    f"Using number of epochs for stage {desc.epochs}"
                )
                desc.epochs_lr = desc.epochs



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import tensorflow as tf

from nncf.common.sparsity.collector import BaseSparseModelStatisticsCollector
from nncf.common.sparsity.collector import WeightDescription
from nncf.tensorflow.graph.utils import get_nncf_operations
from nncf.tensorflow.sparsity.magnitude.functions import apply_mask
from nncf.tensorflow.sparsity.magnitude.operation import BinaryMaskWithWeightsBackup


def _get_standardized_weight_shape(shape):
    return [0 if x is None else x for x in shape]


class TFSparseModelStatisticsCollector(BaseSparseModelStatisticsCollector):
    """
    Collects statistics for the sparse tf.keras.Model.
    """

    def __init__(self, model: tf.keras.Model, operation_names: List[str]):
        """
        Initializes statistics collector of the sparse tf.keras.Model.

        :param model: Sparse model.
        :param operation_names: Names of operations.
        """
        self._model = model
        self._operation_names = operation_names
        self._excluded_names = []
        self._sw_name_to_num_nonzero_map = {}

    def _collect_weights_descriptions(self) -> List[WeightDescription]:
        weights_descriptions = []
        excluded_names = []

        # Collect description for sparse weights i.e. weights for which
        # sparsity algorithm was applied.
        for wrapped_layer, weight_attr, op in get_nncf_operations(self._model, self._operation_names):
            weight = wrapped_layer.layer_weights[weight_attr]
            operation_weights = wrapped_layer.get_operation_weights(op.name)
            binary_mask = op.get_binary_mask(operation_weights)
            sparse_weight = apply_mask(weight, binary_mask)

            weights_descriptions.append(
                WeightDescription(
                    weight.name,
                    _get_standardized_weight_shape(weight.shape.as_list()),
                    tf.math.count_nonzero(sparse_weight).numpy().item(),
                    is_sparse=True,
                )
            )

            # Exclude this name because it has been processed.
            excluded_names.append(weight.name)

            # Exclude these names because they were added to the model
            # by the sparsity algorithm.
            excluded_names.extend([w.name for w in operation_weights.values()])
            if isinstance(op, BinaryMaskWithWeightsBackup):
                excluded_names.append(op.bkup_var.name)

        # Collect descriptions for rest weights.
        unique_weights = {id(w): w for w in self._model.weights if w.name not in excluded_names}.values()
        for weight in unique_weights:
            weights_descriptions.append(
                WeightDescription(
                    weight.name,
                    _get_standardized_weight_shape(weight.shape.as_list()),
                    tf.math.count_nonzero(weight).numpy().item(),
                    is_sparse=False,
                )
            )

            return weights_descriptions



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import List, Optional, TypeVar

from nncf import Dataset
from nncf.common.graph.graph import NNCFGraph
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType

TModel = TypeVar("TModel")


class Algorithm(ABC):
    """
    Base class for all Post-Training algorithms.
    """

    @property
    @abstractmethod
    def available_backends(self) -> List[BackendType]:
        """
        Returns list of the available backends for the algorithm.

        :return: List of backends supported by the algorithm.
        """

    @abstractmethod
    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        """
        Applies the algorithm to the model.

        :param model: Model for applying algorithm.
        :param graph: Model graph.
        :param statistic_points: Statistic points with collected statistics values.
        :param dataset: A representative dataset for the calibration process.
        :return: A resulting model.
        """

    @abstractmethod
    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        """
        Returns statistic points, for which StatisticsCollector should collect statistics.

        :param model: Model for statistics collection.
        :param graph: Model graph.
        :return: Statistic points, for which StatisticsCollector should collect statistics.
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod

from nncf.common.utils.api_marker import api


@api()
class Statistics(ABC):
    """
    Contains a collection of model- or compression-related data and provides a way for its human-readable
    representation.
    """

    @abstractmethod
    def to_str(self) -> str:
        """
        Returns a representation of the statistics as a human-readable string.
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, TypeVar, Union

from nncf.common.factory import NNCFGraphFactory
from nncf.common.factory import StatisticsAggregatorFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.logging import nncf_logger
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.data.dataset import Dataset
from nncf.quantization.algorithms.algorithm import Algorithm

TModel = TypeVar("TModel")
PipelineStep = List[Algorithm]


def collect_statistics(
    containers: Union[StatisticPointsContainer, List[StatisticPointsContainer]],
    model: TModel,
    graph: NNCFGraph,
    dataset: Dataset,
) -> StatisticPointsContainer:
    """
    Utility method for collecting statistics by model.

    :param statistic_points: Statistic points that need to be collected.
    :param model: A model.
    :param graph: A graph assosiated with a model.
    :param dataset: A dataset.
    :return: Collected statistics.
    """
    if not isinstance(containers, list):
        containers = [containers]

    statistics_aggregator = StatisticsAggregatorFactory.create(model, dataset)
    for container in containers:
        statistics_aggregator.register_statistic_points(container)
    statistics_aggregator.collect_statistics(model, graph)

    return statistics_aggregator.statistic_points


class Pipeline:
    """
    A class for creating pipelines that apply algorithms to a model.

    This class is used for creating custom model processing pipelines
    that encapsulate a series of algorithms to be applied to a model
    using a provided dataset.

    A pipeline consists of pipeline steps. Each pipeline step is a
    sequence of Algorithm class instances whose statistic points are
    combined and collected using the model obtained after the previous
    pipeline step. The collected statistic points are used for all
    algorithms in this step.
    """

    def __init__(self, pipeline_steps: List[PipelineStep]):
        """
        :param pipeline_steps: A sequence of pipeline steps to be executed in order.
        """
        self._pipeline_steps = pipeline_steps

    @property
    def pipeline_steps(self) -> List[PipelineStep]:
        """
        Property that defines the sequence of distinct pipeline steps to
        be executed in order.

        :return: A sequence of pipeline steps to be executed in order.
        """
        return self._pipeline_steps

    def run(self, model: TModel, dataset: Dataset) -> TModel:
        """
        Executes the pipeline on the provided model.

        :param model: A model to which pipeline will be applied.
        :param dataset: A dataset that holds the data items for algorithms.
        :return: The updated model after executing the entire pipeline.
        """
        return self.run_from_step(model, dataset)

    def run_step(
        self,
        step_index: int,
        step_statistics: StatisticPointsContainer,
        model: TModel,
        graph: NNCFGraph,
    ) -> TModel:
        """
        Executes a provided pipeline step on the provided model.

        :param step_index: Zero-based index of the pipeline step that should be executed
        :param step_statistics: Statistics required to execute a pipeline step.
        :param model: A model to which a pipeline step will be applied.
        :param graph: A graph assosiated with a model.
        :return: The updated model after executing the pipeline step.
        """
        current_model = model
        current_graph = graph

        pipeline_steps = self._remove_unsupported_algorithms(get_backend(model))
        pipeline_step = pipeline_steps[step_index]
        for algorithm in pipeline_step[:-1]:
            current_model = algorithm.apply(current_model, current_graph, step_statistics)
            current_graph = NNCFGraphFactory.create(current_model)
        current_model = pipeline_step[-1].apply(current_model, current_graph, step_statistics)

        return current_model

    def run_from_step(
        self,
        model: TModel,
        dataset: Dataset,
        graph: Optional[NNCFGraph] = None,
        start_step_index: int = 0,
        step_index_to_statistics: Optional[Dict[int, StatisticPointsContainer]] = None,
    ) -> TModel:
        """
        Executes the pipeline from the specified pipeline step to the end.

        :param model: This is the model after the (start_step_index - 1)-th pipeline
            step, or the initial model if start_step_index is 0.
        :param dataset: A dataset that holds the data items for pipeline steps.
        :param graph: A graph assosiated with a model.
        :param start_step_index: Zero-based pipeline step index from which the pipeline
            should be executed.
        :param step_index_to_statistics: A mapping from pipeline step index to statistics
            required to execute pipeline step.
        :return: The updated model after executing the pipeline from the specified pipeline
            step to the end.
        """
        pipeline_steps = self._remove_unsupported_algorithms(get_backend(model))
        if step_index_to_statistics is None:
            step_index_to_statistics = {}

        # The `step_model` and `step_graph` entities are required to execute `step_index`-th pipeline step
        step_model = model
        step_graph = graph
        for step_index in range(start_step_index, len(pipeline_steps)):
            # Create graph required to run current pipeline step
            if step_graph is None:
                step_graph = NNCFGraphFactory.create(step_model)

            # Collect statistics required to run current pipeline step
            step_statistics = step_index_to_statistics.get(step_index)
            if step_statistics is None:
                statistic_points = self.get_statistic_points_for_step(step_index, step_model, step_graph)
                step_statistics = collect_statistics(statistic_points, step_model, step_graph, dataset)

            # Run current pipeline step
            step_model = self.run_step(step_index, step_statistics, step_model, step_graph)

            step_graph = None  # We should rebuild the graph for the next pipeline step

        return step_model

    def get_statistic_points_for_step(
        self, step_index: int, model: TModel, graph: NNCFGraph
    ) -> StatisticPointsContainer:
        """
        Returns statistics that should be collected to execute `step_index`-th pipeline step.

        :param step_index: Zero-based index of the pipeline step.
        :param model: A model.
        :param graph: A graph assosiated with a model.
        :return: Statistics that should be collected to execute `step_index`-th pipeline step.
        """
        container = StatisticPointsContainer()
        pipeline_steps = self._remove_unsupported_algorithms(get_backend(model))
        pipeline_step = pipeline_steps[step_index]
        for algorithm in pipeline_step:
            for statistic_points in algorithm.get_statistic_points(model, graph).values():
                for statistic_point in statistic_points:
                    container.add_statistic_point(statistic_point)

        return container

    def _remove_unsupported_algorithms(self, backend: BackendType) -> List[PipelineStep]:
        pipeline_steps = []
        for pipeline_step in self._pipeline_steps:
            step = []
            for algorithm in pipeline_step:
                if backend not in algorithm.available_backends:
                    nncf_logger.debug(f"{backend.name} does not support {algorithm.__class__.__name__} algorithm yet.")
                    continue
                step.append(algorithm)

            if step:
                pipeline_steps.append(step)

        return pipeline_steps



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import operator
from collections import OrderedDict
from collections import defaultdict
from functools import partial
from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Tuple, TypeVar

import nncf
from nncf import Dataset
from nncf.common.factory import StatisticsAggregatorFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.logging import nncf_logger
from nncf.common.logging.track_progress import track
from nncf.common.scopes import should_consider_scope
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.common.utils.helpers import create_table
from nncf.experimental.common.tensor_statistics.statistics import WCTensorStatistic
from nncf.parameters import BackupMode
from nncf.parameters import CompressWeightsMode
from nncf.parameters import SensitivityMetric
from nncf.quantization.advanced_parameters import AdvancedCompressionParameters
from nncf.quantization.advanced_parameters import convert_to_dict_recursively
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.quantization.algorithms.weight_compression.awq import AWQ
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.gptq import GPTQ
from nncf.quantization.algorithms.weight_compression.lora_correction import LoraCorrectionAlgorithm
from nncf.quantization.algorithms.weight_compression.mixed_precision import MIXED_PRECISION_CRITERIA
from nncf.quantization.algorithms.weight_compression.scale_estimation import ScaleEstimation
from nncf.quantization.algorithms.weight_compression.weight_lowering import WeightCompressionConfig
from nncf.scopes import IgnoredScope
from nncf.scopes import get_ignored_node_names_from_ignored_scope

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")

INT8_MODES = [CompressWeightsMode.INT8_ASYM, CompressWeightsMode.INT8_SYM]
NON_INT8_MODES = [
    CompressWeightsMode.INT4_SYM,
    CompressWeightsMode.INT4_ASYM,
    CompressWeightsMode.NF4,
    CompressWeightsMode.E2M1,
]


def get_weight_compression_configuration(
    mode: CompressWeightsMode = CompressWeightsMode.INT8_ASYM,
    dataset: Optional[Dataset] = None,
    ratio: Optional[float] = None,
    group_size: Optional[int] = None,
    all_layers: Optional[bool] = None,
    awq: Optional[bool] = None,
    scale_estimation: Optional[bool] = None,
    gptq: Optional[bool] = None,
    lora_correction: Optional[bool] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    sensitivity_metric: Optional[SensitivityMetric] = None,
    backup_mode: Optional[BackupMode] = None,
    advanced_parameters: Optional[AdvancedCompressionParameters] = None,
) -> Dict[str, Any]:
    """
    Generates a configuration dictionary for weight compression based on the provided parameters.
    """
    group_size = (
        -1
        if group_size is None and mode in INT8_MODES
        else 128 if group_size is None and mode in NON_INT8_MODES else group_size
    )

    return {
        "mode": mode,
        "ratio": ratio or 1,
        "group_size": group_size,
        "all_layers": all_layers or False,
        "awq": awq or False,
        "scale_estimation": scale_estimation or False,
        "gptq": gptq or False,
        "lora_correction": lora_correction or False,
        "ignored_scope": ignored_scope or IgnoredScope(),
        "sensitivity_metric": (
            (
                SensitivityMetric.WEIGHT_QUANTIZATION_ERROR
                if dataset is None
                else SensitivityMetric.MAX_ACTIVATION_VARIANCE
            )
            if sensitivity_metric is None
            else sensitivity_metric
        ),
        "backup_mode": backup_mode or BackupMode.INT8_ASYM,
        "advanced_parameters": advanced_parameters or AdvancedCompressionParameters(),
    }


def check_user_compression_configuration(
    mode: CompressWeightsMode,
    subset_size: int,
    dataset: Optional[Dataset],
    ratio: Optional[float],
    group_size: Optional[int],
    all_layers: Optional[bool],
    awq: Optional[bool],
    scale_estimation: Optional[bool],
    gptq: Optional[bool],
    lora_correction: Optional[bool],
    ignored_scope: Optional[IgnoredScope],
    sensitivity_metric: Optional[SensitivityMetric],
    backup_mode: Optional[BackupMode],
    advanced_parameters: Optional[AdvancedCompressionParameters],
) -> None:
    """
    Validates the user's weight compression configuration for correctness.
    """
    if mode in INT8_MODES:
        if (ratio and ratio != 1) or (group_size and group_size != -1):
            raise nncf.ParameterNotSupportedError(
                "INT8 modes require per-channel quantization of all layers in 8 bit. "
                "Default values of `ratio` (1) and `group_size` (-1) cannot be overridden."
            )

        if advanced_parameters and advanced_parameters.statistics_path:
            raise nncf.ParameterNotSupportedError(
                "INT8 modes do not support the `statistics_path` option in `AdvancedCompressionParameters`."
            )

        unsupported_options = {
            "all_layers": all_layers,
            "sensitivity_metric": sensitivity_metric,
            "dataset": dataset,
            "awq": awq,
            "scale_estimation": scale_estimation,
            "gptq": gptq,
            "lora_correction": lora_correction,
            "backup_mode": backup_mode,
        }
        unsupported_for_int8 = [name for name, value in unsupported_options.items() if value is not None]
        if unsupported_for_int8:
            raise nncf.ParameterNotSupportedError(
                f"INT8 modes do not support {', '.join(unsupported_for_int8)} option(s). Set them to None."
            )

    if ratio is not None and not (0 <= ratio <= 1):
        raise nncf.ValidationError(f"The ratio should be between 0 and 1, but ratio={ratio} is specified.")

    if subset_size <= 0:
        raise nncf.ValidationError(f"The subset_size value should be positive, but subset_size={subset_size} is given.")

    if (
        ratio
        and dataset is None
        and sensitivity_metric is not None
        and sensitivity_metric != SensitivityMetric.WEIGHT_QUANTIZATION_ERROR
    ):
        raise nncf.ValidationError(
            f"Mixed precision selection with sensitivity metric={sensitivity_metric.value} \
            requires a dataset, but it's not provided."
        )


class WeightCompression(Algorithm):
    """
    Post-training Weight Compression algorithm implementation.

    Compresses weights of Linear and Embedding layers to 8-bit integer or
    to 4-bit integer/float depending on mode, ratio and group size.
    """

    def __init__(
        self,
        mode: CompressWeightsMode,
        ratio: float,
        group_size: int,
        ignored_scope: IgnoredScope,
        all_layers: bool,
        sensitivity_metric: SensitivityMetric,
        awq: bool,
        subset_size: int,
        scale_estimation: bool,
        gptq: bool,
        lora_correction: bool,
        backup_mode: BackupMode = BackupMode.INT8_ASYM,
        advanced_parameters: Optional[AdvancedCompressionParameters] = None,
    ):
        """
        :param mode: Defines a mode for weight compression.
            INT8_SYM stands for 8-bit integer symmetric quantization of all weights.
                Weights are quantized symmetrically without zero point.
            INT8_ASYM is the same as INT8_SYM mode, but weights are quantized to a primary precision asymmetrically
                with a typical non-fixed zero point.
            INT4_SYM stands for a mixed-precision weights quantization with 4-bit integer as a primary precision.
                Weights are quantized to a primary precision symmetrically without zero point.
                All embeddings and the last layer are always compressed to a backup_mode, which is INT8_ASYM,
                by default. All others are quantized whether to 4-bit integer or to a backup_mode depending on
                criteria and the given ratio.
            INT4_ASYM is the same as INT4_SYM mode, but weights are quantized to a primary precision asymmetrically
                with a typical non-fixed zero point.
            NF4 is the same as INT4_SYM mode, but primary precision is NF4 data type without zero point.
            E2M1 is the same as INT4_SYM mode, but primary precision is E2M1 data type without zero point.
        :param ratio: the ratio between primary and backup precisions (e.g. 0.9 means 90% of layers quantized to NF4
            and the rest to backup_mode).
        :param group_size: number of weights (e.g. 128) in the channel dimension
            that share quantization parameters (scale). The value -1 means no grouping.
        :param ignored_scope: An ignored scope that defined the list of model control
            flow graph nodes to be ignored during quantization.
        :param all_layers: Indicates whether embeddings and last MatMul layers should be compressed to a primary
            precision. By default, the backup precision is assigned for the embeddings and last MatMul layers.
        :param sensitivity_metric: The sensitivity metric for assigning quantization precision to layers. In order to
            preserve the accuracy of the model, the more sensitive layers receives a higher precision.
        :param awq: determines whether to use or not modified AWQ algorithm.
        :param subset_size: Number of data samples to calculate activation statistics used for assigning different
            quantization precision.
        :param scale_estimation: determines whether to use or not scale estimation for 4 bit layers.
        :param gptq: determines whether to use or not GPTQ algorithm.
        :param lora_correction: determines whether to use or not LoRA Correction algorithm.
        :param backup_mode: Defines a backup mode for mixed-precision weight compression.
            NONE stands for original floating-point precision of the model weights.
                In this mode, weights are retained in their original precision without any quantization.
            INT8_SYM stands for 8-bit integer symmetric quantization without zero point.
            INT8_ASYM stands for 8-bit integer asymmetric quantization with a typical non-fixed zero point.
        :param advanced_parameters: advanced parameters for algorithms in compression pipeline.
        """
        super().__init__()
        self._mode = mode
        self._group_size = group_size
        self._ratio = ratio
        self._ignored_scope = ignored_scope
        self._backend_entity = None
        self._algorithm_key = f"CW_{hash(self)}"
        self._statistics = {}
        self._all_layers = all_layers
        self._sensitivity_metric = sensitivity_metric
        self._awq = awq
        self._subset_size = subset_size
        self._scale_estimation = scale_estimation
        self._gptq = gptq
        self._lora_correction = lora_correction
        self._backup_mode = backup_mode
        self._advanced_parameters = (
            advanced_parameters if advanced_parameters is not None else AdvancedCompressionParameters()
        )

        primary_config = WeightCompressionConfig(mode=self._mode, group_size=self._group_size)
        criterion_cls = MIXED_PRECISION_CRITERIA.get(self._sensitivity_metric)
        self._mixed_precision_algo = criterion_cls(primary_config, self._ratio, self._subset_size)
        self._statistics_path = self._advanced_parameters.statistics_path
        if self._gptq:
            gptq_params = self._advanced_parameters.gptq_params
            self._gptq_algo = GPTQ(
                damp_percent=gptq_params.damp_percent,
                block_size=gptq_params.block_size,
                subset_size=gptq_params.subset_size,
                scale_estimation=self._scale_estimation,
            )

        self._data_aware_mixed_precision = (
            self._sensitivity_metric != SensitivityMetric.WEIGHT_QUANTIZATION_ERROR and self._ratio != 1.0
        )
        self._data_aware_compression = self._awq or self._scale_estimation or self._lora_correction or self._gptq

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO, BackendType.TORCH, BackendType.TORCH_FX]

    def set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVWeightCompressionAlgoBackend

            self._backend_entity = OVWeightCompressionAlgoBackend(model)
        elif model_backend == BackendType.TORCH:
            from nncf.quantization.algorithms.weight_compression.torch_backend import PTWeightCompressionAlgoBackend

            self._backend_entity = PTWeightCompressionAlgoBackend()
        elif model_backend == BackendType.TORCH_FX:
            from nncf.quantization.algorithms.weight_compression.torch_fx_backend import FXWeightCompressionAlgoBackend

            self._backend_entity = FXWeightCompressionAlgoBackend()
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def get_nodes_to_compress(self, nncf_graph: NNCFGraph) -> List[NNCFNode]:
        """
        Collects nodes in the model's graph corresponding to the layers for weight compression.

        :param nncf_graph: NNCFGraph instance.
        :return: List with the data for each layer.
        """
        weighted_metatypes = (
            self._backend_entity.matmul_metatypes
            + self._backend_entity.embedding_metatypes
            + self._backend_entity.convolution_metatypes
        )

        ordered_nodes_to_compress = []
        ignored_names = get_ignored_node_names_from_ignored_scope(
            self._ignored_scope, nncf_graph, strict=self._ignored_scope.validate
        )
        for node in nncf_graph.topological_sort():
            is_node_with_weights = self._backend_entity.is_node_with_weights(node, nncf_graph)
            is_within_scope = should_consider_scope(node.node_name, ignored_names)
            if node.metatype in weighted_metatypes and is_node_with_weights and is_within_scope:
                ordered_nodes_to_compress.append(node)
        return ordered_nodes_to_compress

    def _get_ratio_defining_params(
        self, all_weight_params: List[WeightCompressionParameters], is_last_layer_shared: bool
    ) -> List[WeightCompressionParameters]:
        """
        Returns the information about weights that are used for ratio calculation between primary
        and backup precisions.

        :param all_weight_params: List of all weight parameters.
        :param is_last_layer_shared: Indicates whether the last layer which shares the weight
            should be quantized or not.
        :return: Information about each weight node that is considered for mixed precision.
        """
        if self._mode in [CompressWeightsMode.INT8_SYM, CompressWeightsMode.INT8_ASYM]:
            return all_weight_params

        ratio_defining_params = list(
            filter(
                lambda wp: wp.node_with_weight.metatype in self._backend_entity.matmul_metatypes,
                all_weight_params,
            )
        )

        # The last MatMul layer is quantized to 4-bits if all_layers=True
        if not self._all_layers and not is_last_layer_shared:
            ratio_defining_params = ratio_defining_params[:-1]

        # Embedding layers are quantized to 4-bits only if all_layers=True.
        if self._all_layers:
            embedding_params = list(
                filter(
                    lambda wp: wp.node_with_weight.metatype in self._backend_entity.embedding_metatypes
                    and len(wp.reduction_axes) == 1,
                    all_weight_params,
                )
            )
            ratio_defining_params.extend(embedding_params)

        return ratio_defining_params

    def _set_weight_compression_config(
        self,
        ratio_defining_params: List[WeightCompressionParameters],
        model: TModel,
        graph: NNCFGraph,
        statistics_points: StatisticPointsContainer,
    ) -> None:
        """
        Sets the appropriate compression configuration for weights based on some criteria.

        :param ratio_defining_params: Information about weights that are used for calculating ratio between primary and
            backup precisions.
        :param model: The model.
        :param graph: The model graph associated with the model.
        :param statistics_points: Statistics points.
        """
        primary_config = WeightCompressionConfig(mode=self._mode, group_size=self._group_size)
        if self._ratio == 1:
            for weight_param in ratio_defining_params:
                weight_param.compression_config = primary_config
        else:
            self._mixed_precision_algo.apply(model, graph, statistics_points, weight_params=ratio_defining_params)

    @staticmethod
    def _proportion_str(num_weights_list: List[int], total_num_weights: int, total_num_params: int) -> str:
        """
        Generates a string with proportion between target parameters and all model parameters by number of weights.

        :param num_weights_list: List of number of weights of target model parameters.
        :param total_num_weights: The total number of weights.
        :param total_num_params: The total number of model parameters.
        :return: The string with proportion between target parameters and all model parameters by number of weights.
        """
        percentage = sum(num_weights_list) / max(total_num_weights, 1) * 100
        return f"{percentage:.0f}% ({len(num_weights_list)} / {total_num_params})"

    def _get_bitwidth_distribution_str(
        self,
        all_params: List[WeightCompressionParameters],
        ratio_defining_params: List[WeightCompressionParameters],
        ignored_scope_weight_statistics: List[int],
    ) -> str:
        """
        Generates a table that shows the ratio of weights quantized to different number of bits.

        :param all_params: Information about each weight node.
        :param ratio_defining_params: Information about weights that are used for calculating ratio between primary and
            backup precisions.
        :param ignored_scope_weight_statistics: Information about weight nodes from IgnoredScope.
        :return: A string containing the table.
        """
        dtype_vs_num_weights_map = {}
        ratio_defining_weight_names = set(wp.weight_name for wp in ratio_defining_params)
        for data in all_params:
            dtype = data.compression_config.mode if data.compression_config is not None else "float"
            n_total, n_ratio_defining = dtype_vs_num_weights_map.get(dtype, ([], []))
            if data.weight_name in ratio_defining_weight_names:
                n_ratio_defining.append(data.num_weights)
            n_total.append(data.num_weights)
            dtype_vs_num_weights_map[dtype] = (n_total, n_ratio_defining)

        if ignored_scope_weight_statistics:
            n_total, n_ratio_defining = dtype_vs_num_weights_map.get("float", ([], []))
            dtype_vs_num_weights_map["float"] = (n_total + ignored_scope_weight_statistics, n_ratio_defining)

        num_ratio_defining_weights = sum(ws.num_weights for ws in ratio_defining_params)
        num_ratio_defining_params = len(ratio_defining_params)
        num_total_weights = sum(ws.num_weights for ws in all_params) + sum(ignored_scope_weight_statistics)
        num_params = len(all_params) + len(ignored_scope_weight_statistics)
        dtype_vs_num_weights_map = OrderedDict(sorted(dtype_vs_num_weights_map.items(), reverse=True))
        # Table creation
        header = ["Weight compression mode", "% all parameters (layers)", "% ratio-defining parameters (layers)"]
        rows = []
        for bitwidth, (n_total, n_ratio_defining) in dtype_vs_num_weights_map.items():
            rows.append(
                [
                    bitwidth,
                    self._proportion_str(n_total, num_total_weights, num_params),
                    self._proportion_str(n_ratio_defining, num_ratio_defining_weights, num_ratio_defining_params),
                ]
            )

        table = create_table(header, rows)
        pretty_string = f"Statistics of the bitwidth distribution:\n{table}"
        return pretty_string

    def _get_ignored_scope_weight_statistics(self, model: TModel, graph: NNCFGraph) -> List[int]:
        """
        Collect the weight statistics for nodes in the ignored scope.

        :param model: Model for statistics collection.
        :param graph: Model graph.
        :return: A list of weight sizes for the ignored nodes.
        """
        ignored_names = get_ignored_node_names_from_ignored_scope(self._ignored_scope, graph, strict=False)
        weighted_metatypes = (
            self._backend_entity.matmul_metatypes
            + self._backend_entity.embedding_metatypes
            + self._backend_entity.convolution_metatypes
        )
        ignored_scope_weight_statistics = []
        for node_name in ignored_names:
            node = graph.get_node_by_name(node_name)
            is_node_with_weights = self._backend_entity.is_node_with_weights(node, graph)
            if not is_node_with_weights or node.metatype not in weighted_metatypes:
                continue
            for _, weight_port_id in self._backend_entity.get_weight_names_and_port_ids(node, graph):
                weight_dtype = self._backend_entity.get_weight_dtype(node, weight_port_id, model, graph)
                if weight_dtype.is_float():
                    continue
                weight_shape = self._backend_entity.get_weight_shape(node, weight_port_id, graph)
                weight_size = reduce(operator.mul, weight_shape, 1)
                ignored_scope_weight_statistics.append(weight_size)
        return ignored_scope_weight_statistics

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        self.set_backend_entity(model)

        nodes_to_compress = self.get_nodes_to_compress(graph)

        statistics = None
        if self._data_aware_mixed_precision or self._data_aware_compression:
            matmul_nodes_to_compress = [
                node for node in nodes_to_compress if node.metatype in self._backend_entity.matmul_metatypes
            ]
            matmul_input_to_output_nodes_map = self.get_matmul_input_to_output_nodes_map(
                matmul_nodes_to_compress, graph
            )
            if statistic_points is None:
                statistic_points = self.get_statistic_points(model, graph, matmul_input_to_output_nodes_map.keys())
                statistic_points = self._collect_statistics(dataset, graph, model, statistic_points)
            statistics = self._get_statistics_for_weights_compression(
                matmul_input_to_output_nodes_map, statistic_points
            )

        all_weight_params: List[WeightCompressionParameters] = []
        weight_names = set()

        is_last_layer_shared = False
        n = len(nodes_to_compress)
        for i, node in enumerate(nodes_to_compress):
            for weight_name, weight_port_id in self._backend_entity.get_weight_names_and_port_ids(node, graph):
                if weight_name in weight_names:
                    if i == n - 1:
                        is_last_layer_shared = True
                    continue

                weight_dtype = self._backend_entity.get_weight_dtype(node, weight_port_id, model, graph)
                if not weight_dtype.is_float():
                    continue
                weight_shape = self._backend_entity.get_weight_shape(node, weight_port_id, graph)
                weight_size = reduce(operator.mul, weight_shape, 1)
                reduction_axes = self._backend_entity.get_reduction_axes(node, weight_port_id, graph)
                if (
                    self._group_size != -1
                    and self._all_layers
                    and node.metatype in self._backend_entity.embedding_metatypes
                    and isinstance(reduction_axes, tuple)
                    and len(reduction_axes) != 1
                ):
                    # NNCF supports multiple reduction axes only for ops with group_size != -1.
                    # Convolution ops are always kept in backup mode.
                    # Embedding layers are quantized to 4-bits only if all_layers=True.
                    # MatMul ops can't have multiple reduction axes.
                    nncf_logger.warning(
                        f"Weight compression expects a single reduction axis, but {len(reduction_axes)} given. "
                        f"Weight shape: {weight_shape}, reduction axes: {reduction_axes}, "
                        f"node name: {node.node_name}. The node will be in {self._backup_mode} mode."
                    )

                if self._backup_mode == BackupMode.NONE:
                    wc_config = None
                else:
                    mode = (
                        CompressWeightsMode.INT8_ASYM
                        if self._backup_mode == BackupMode.INT8_ASYM
                        else CompressWeightsMode.INT8_SYM
                    )
                    wc_config = WeightCompressionConfig(mode=mode)
                weight_params = WeightCompressionParameters(
                    weight_name, node, weight_port_id, weight_size, reduction_axes, wc_config
                )
                all_weight_params.append(weight_params)
                weight_names.add(weight_name)

        ratio_defining_params = self._get_ratio_defining_params(all_weight_params, is_last_layer_shared)
        self._set_weight_compression_config(ratio_defining_params, model, graph, statistic_points)
        ignored_scope_weight_statistics = self._get_ignored_scope_weight_statistics(model, graph)
        nncf_logger.info(
            self._get_bitwidth_distribution_str(
                all_weight_params, ratio_defining_params, ignored_scope_weight_statistics
            )
        )

        if self._backup_mode == BackupMode.NONE:
            # Filter all_weight_params and nodes_to_compress by excluding nodes
            # that should remain in their original floating-point precision
            nodes_names_to_exclude = {
                w_params.node_with_weight.node_name
                for w_params in all_weight_params
                if w_params.compression_config is None
            }
            all_weight_params = list(
                filter(
                    lambda w_params: w_params.node_with_weight.node_name not in nodes_names_to_exclude,
                    all_weight_params,
                )
            )
            nodes_to_compress = list(
                filter(lambda node: node.node_name not in nodes_names_to_exclude, nodes_to_compress)
            )

        if self._awq:
            awq_params = self._advanced_parameters.awq_params
            awq_algo = AWQ(
                model,
                self._backend_entity.name_to_node_mapping,
                all_weight_params,
                nodes_to_compress,
                statistics,
                awq_params.subset_size,
                awq_params.percent_to_apply,
                awq_params.alpha_min,
                awq_params.alpha_max,
                awq_params.steps,
            )
            awq_algo.apply(model, graph)
            # After applying AWQ we need to update statistics since AWQ alters the activations
            statistics = awq_algo.update_statistics(statistics)
            # del is used to prematurely mark non-necessary data as free for garbage collection
            del awq_algo

        scales = {}
        zero_points = {}
        lora_correction_algo = None
        description = "Applying Weight Compression"
        if self._gptq:
            del statistics
            model, scales, zero_points = self._gptq_algo.apply(
                model=model,
                graph=graph,
                dataset=dataset,
                weight_compression_parameters=all_weight_params,
                backend_entity=self._backend_entity,
            )
        else:
            if self._scale_estimation:
                scale_estimation_params = self._advanced_parameters.scale_estimation_params
                scales, zero_points = ScaleEstimation(
                    model,
                    self._backend_entity.name_to_node_mapping,
                    all_weight_params,
                    nodes_to_compress,
                    statistics,
                    scale_estimation_params.subset_size,
                    scale_estimation_params.initial_steps,
                    scale_estimation_params.scale_steps,
                    scale_estimation_params.weight_penalty,
                ).apply(model, graph)

            if self._lora_correction:
                lora_correction_params = self._advanced_parameters.lora_correction_params
                lora_correction_algo = LoraCorrectionAlgorithm(statistics, lora_correction_params)
                description += " with correction of low-rank adapters"
            del statistics

        # Sort weight params to start compression with the bigger constants. This lowers peak memory footprint.
        all_weight_params = sorted(all_weight_params, key=lambda wp: wp.num_weights, reverse=True)
        all_weight_sizes = [wp.num_weights for wp in all_weight_params]

        # Compress model using weight compression parameters
        transformed_model = self._backend_entity.transform_model(
            model,
            graph,
            track(all_weight_params, description=description, weights=all_weight_sizes),
            scales,
            zero_points,
            lora_correction_algo,
        )

        self._backend_entity.dump_parameters(
            model,
            parameters={
                "mode": self._mode.value,
                "group_size": self._group_size,
                "ratio": self._ratio,
                "all_layers": self._all_layers,
                "ignored_scope": self._ignored_scope,
                "sensitivity_metric": self._sensitivity_metric.value,
                "awq": self._awq,
                "scale_estimation": self._scale_estimation,
                "gptq": self._gptq,
                "lora_correction": self._lora_correction,
                "backup_mode": self._backup_mode.value,
                "advanced_parameters": convert_to_dict_recursively(self._advanced_parameters),
            },
            algo_name="weight_compression",
        )
        return transformed_model

    def _get_activation_node_and_port(self, node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[NNCFNode, int]:
        """
        This method returns the activation layer and corresponding port id for the node.

        :param node: NNCFGraph node for which the activation is sought.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Tuple with the activation node and port id.
        """
        activation_port = self._backend_entity.get_activation_port_id(node, nncf_graph)
        activation_edge = nncf_graph.get_input_edge_by_port_id(node, activation_port)
        activation_node = activation_edge.from_node
        port_id = activation_edge.output_port_id
        return activation_node, port_id

    def get_matmul_input_to_output_nodes_map(
        self, matmul_nodes: List[NNCFNode], graph: NNCFGraph
    ) -> Dict[Tuple[NNCFNode, int], List[NNCFNode]]:
        """
        Maps activation nodes to their corresponding MatMul nodes in the graph.

        Each weighted MatMul node takes two inputs: an activation and a weight.
        An activation node may serve as an input to multiple MatMul nodes.
        This function returns a mapping where each key is a tuple consisting of an
        activation node and its output port ID, and the value is a list of MatMul
        nodes that use this activation as input.

        :param matmul_nodes: A list of MatMul nodes from the computation graph.
        :param graph: An instance of NNCFGraph representing the computation graph.
        :return: A dictionary mapping from a tuple of (activation node, port ID)
        to a list of corresponding MatMul nodes that accept the activation as input.
        """
        matmul_input_to_output_nodes_map = defaultdict(list)
        for node in matmul_nodes:
            if node.layer_attributes.input_attributes["transpose"]:  # It works only for OV
                raise nncf.UnsupportedModelError("Transposed input is not supported")
            act_node, output_port_id = self._get_activation_node_and_port(node, graph)
            matmul_input_to_output_nodes_map[(act_node, output_port_id)].append(node)
        return matmul_input_to_output_nodes_map

    def get_compression_nodes_info(
        self, graph: NNCFGraph
    ) -> Tuple[List[NNCFNode], Dict[Tuple[NNCFNode, int], List[NNCFNode]]]:
        """
        Retrieves the nodes to compress along with a mapping of activation nodes
        to their corresponding MatMul nodes.

        This function first identifies all nodes that can be compressed from the
        provided graph. It then filters these nodes to find those that are of
        MatMul type and generates a mapping of activation nodes to their
        corresponding MatMul nodes using the
        `get_matmul_input_to_output_nodes_map` function.

        :param graph: An instance of NNCFGraph representing the computation graph.
        :return: A tuple containing:
        - Nodes for compression.
        - A dictionary mapping from a tuple of (activation node, port ID)
        to a list of MatMul nodes that accept the activation as input.
        """
        nodes_to_compress = self.get_nodes_to_compress(graph)
        matmul_nodes_to_compress = [
            node for node in nodes_to_compress if node.metatype in self._backend_entity.matmul_metatypes
        ]
        matmul_input_to_output_nodes_map = self.get_matmul_input_to_output_nodes_map(matmul_nodes_to_compress, graph)
        return nodes_to_compress, matmul_input_to_output_nodes_map

    def _collect_statistics(
        self,
        dataset: Dataset,
        graph: NNCFGraph,
        model: TModel,
        statistic_points: StatisticPointsContainer,
    ):
        """
        Creates statistics aggregator, registers all statistics specified for algorithm, and then collect them.

        :param dataset: Dataset to collect values.
        :param graph: Model graph.
        :param model: Model for statistics collection.
        :param statistic_points: Statistics points.
        """
        statistics_aggregator = StatisticsAggregatorFactory.create(model, dataset)
        statistics_aggregator.register_statistic_points(statistic_points)
        statistics_aggregator.collect_statistics(model, graph)
        return statistics_aggregator.statistic_points

    def get_statistic_points(
        self,
        model: TModel,
        graph: NNCFGraph,
        nodes_and_port_ids: Iterable[Tuple[NNCFNode, int]],
    ) -> StatisticPointsContainer:
        """
        Returns statistic points, for which StatisticsCollector should collect statistics.

        :param model: Model for statistics collection.
        :param graph: Model graph.
        :param nodes_and_port_ids: Nodes and port ids for which statistics should be collected.
        :return: Statistic points, for which StatisticsCollector should collect statistics.
        """
        statistic_container = StatisticPointsContainer()
        # Statistics for data aware algorithms
        if self._data_aware_compression:
            for node, output_port_id in nodes_and_port_ids:
                statistic_point = self._backend_entity.target_point(
                    TargetType.POST_LAYER_OPERATION, node.node_name, port_id=output_port_id
                )
                # Reduce activations across all but the last dimension. The last dimension is assumed to be the hidden
                # size dimension.
                n_dims = len(graph.get_output_edges_by_port_id(node, output_port_id)[0].tensor_shape)
                stat_collector = self._backend_entity.mean_statistic_collector(
                    reduction_axes=tuple(range(n_dims - 1)), subset_size=self._subset_size
                )
                statistic_container.add_statistic_point(
                    StatisticPoint(
                        target_point=statistic_point, tensor_collector=stat_collector, algorithm=self._algorithm_key
                    )
                )
        # Statistics for mixed precision algorithm
        if self._data_aware_mixed_precision:
            mixed_precision_statistics = self._mixed_precision_algo.get_statistic_points(
                model, graph, nodes_and_port_ids
            )
            for points in mixed_precision_statistics.values():
                for point in points:
                    statistic_container.add_statistic_point(point)

        return statistic_container

    def _get_statistics_for_weights_compression(
        self,
        matmul_input_to_output_nodes_map: Dict[Tuple[NNCFNode, int], List[NNCFNode]],
        statistic_points: StatisticPointsContainer,
    ) -> Dict[str, WCTensorStatistic]:
        """
        Retrieve collected statistics only for WeightCompression algorithm and not for MixedPrecision.

        :param matmul_input_to_output_nodes_map: A mapping from activation node and a port id to corresponding matmul
            nodes which accept this activation as an input.
        :param statistic_points: Statistic points object.
        :return: Collected statistics.
        """

        def input_filter_func(point, port_id):
            # For the floating-point statistics collected in POST_LAYER style,
            # we also need to determine the output port id.
            # For the cases when the layer has more than one (0) output port.
            return (
                self._algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == TargetType.POST_LAYER_OPERATION
                and point.target_point.port_id == port_id
            )

        # For each node we store statistics in a WCTensorStatistics data-class. It contains the following fields:
        #   mean_values=[mean_value_1, ..., mean_value_n]
        #   shapes=[shape_1, ..., shape_n]
        # Where mean_value is a 1D tensor representing an activation reduced over batch and sequence length dimensions,
        # shape is an original shape of an activation before reduction, n is the size of the dataset (or subset_size).
        statistics = {}
        for (act_node, output_port_id), matmul_nodes in matmul_input_to_output_nodes_map.items():
            tensor_collectors = list(
                statistic_points.get_algo_statistics_for_node(
                    act_node.node_name, partial(input_filter_func, port_id=output_port_id), self._algorithm_key
                )
            )
            # Statistics could be empty in case when the statistics is registered for another algorithm,
            # e.g. mixed precision.
            if tensor_collectors:
                assert len(tensor_collectors) == 1
                stats = tensor_collectors[0].get_statistics()

                # Each activation node may have multiple MatMul nodes which it is an input to
                for node in matmul_nodes:
                    statistics[node.node_name] = copy.deepcopy(stats)
        return statistics



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from typing import Any, Dict, List, Optional

import tensorflow as tf
from tensorflow.python.eager import context
from tensorflow.python.framework import ops
from tensorflow.python.ops import math_ops

from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.tensorflow.context import get_current_context
from nncf.experimental.tensorflow.graph.argprovider import TF_ARG_PROVIDERS
from nncf.experimental.tensorflow.graph.argprovider import replace_value_by_index
from nncf.experimental.tensorflow.graph.transformations.commands import TFTargetPoint
from nncf.experimental.tensorflow.scope import get_op_name
from nncf.tensorflow.layers.operation import NNCFOperation


class Hook:
    """
    Contains the NNCF operations and target point where
    these operations should be applied.
    """

    def __init__(self, operations: List[NNCFOperation], target_point: TFTargetPoint, ops_weights: Dict[str, Any]):
        """
        Initializes the hook.

        :param operations: List of the NNCF operations in the correct order.
            The operation at index 0 is applied first, with index -1 last.
        :param target_point: A target point. Contains information on where the
            operations should be applied.
        :param ops_weights: Weights of the operations.
        """
        self._operations = operations
        self._target_point = target_point
        self._ops_weights = ops_weights

        arg_provider_cls = TF_ARG_PROVIDERS.registry_dict.get(self._target_point.op_type_name)
        if arg_provider_cls is None:
            raise ValueError(
                f"Unexpected type of the TensorFlow operation: {self._target_point.op_type_name}. "
                "Register an `ArgProvider` instance for this type in the "
                "`TF_ARG_PROVIDERS` registry, please."
            )

        self._arg_provider = arg_provider_cls()

    @property
    def operations(self) -> List[NNCFOperation]:
        return self._operations

    @property
    def target_point(self) -> TFTargetPoint:
        return self._target_point

    @property
    def is_pre_hook(self) -> bool:
        """
        Returns the boolean flag that specified whether the hook should be
        applied before or after target operation.

        :return: `True` if hook should be applied before target operation,
            `False` otherwise.
        """
        return self._target_point.type == TargetType.OPERATOR_PRE_HOOK

    def get_operation_weights(self, op_name: str) -> Any:
        """
        Returns weights of the operation with `op_name`.

        :param op_name: Name of the operation.
        :return: Weihts of the operation.
        """
        return self._ops_weights[op_name]

    def __call__(self, *args, **kwargs):
        """
        Applies this hook.

        :return: A tuple (args, kwargs)
        """
        if self.is_pre_hook:
            get_fn = self._arg_provider.get_input
            set_fn = self._arg_provider.set_input
        else:
            get_fn = self._arg_provider.get_output
            set_fn = self._arg_provider.set_output

        x = get_fn(self.target_point.port_id, args, kwargs)
        for op in self.operations:
            w = self.get_operation_weights(op.name)
            x = op(*(x, w, None))
        args_, kwargs_ = set_fn(self.target_point.port_id, x, args, kwargs)

        return args_, kwargs_


class TensorFlowOpWrapper:
    """
    Describes a wrapper around the method from the TensorFlow API.
    """

    def __init__(self, op, op_type_name: str):
        """
        Initializes a wrapper.

        :param op: Original method.
        :param op_type_name: Operation type name (name of the function
            from the `tf.raw_ops` module).
        """
        self._op = op
        self._op_type_name = op_type_name

    def __call__(self, *args, **kwargs):
        """
        Applies TensorFlow operation with compression extensions.
        """
        tracing_context = get_current_context()

        # Should we wrap current operation?
        if not tracing_context.wrap_ops:
            return self._op(*args, **kwargs)

        op_name = get_op_name(self._op_type_name, kwargs.get("name"))

        _pre_hooks = getattr(get_current_context().model, "_pre_hooks")
        _post_hooks = getattr(get_current_context().model, "_post_hooks")

        with tracing_context.enter(in_call=True, wrap_ops=False):
            # Apply pre-hooks
            args, kwargs = TensorFlowOpWrapper._apply_hooks(_pre_hooks.get(op_name, []), args, kwargs)

            # Apply TensorFlow operation
            outputs = self._op(*args, **kwargs)

            # Apply post-hooks
            (outputs,), _ = TensorFlowOpWrapper._apply_hooks(_post_hooks.get(op_name, []), (outputs,), {})

        return outputs

    def __getattr__(self, name):
        return getattr(self._op, name)

    @staticmethod
    def _apply_hooks(hooks: List[Hook], args, kwargs):
        """
        Applies hooks.

        :param hooks: A list of hooks.
        :return: A tuple (args, kwargs).
        """
        for hook in hooks:
            args, kwargs = hook(*args, **kwargs)
        return args, kwargs


class TFPatcher:
    """
    Performs modifications of the TensorFlow code.
    """

    @staticmethod
    def patch_tf_operations() -> None:
        """
        Applies patches to the TensorFlow code.
        """
        op_type_name_to_op_info_map = TFPatcher._get_ops_info()
        for op_type_name, (fn, fn_name, module) in op_type_name_to_op_info_map.items():
            tf_op_wrapper = TensorFlowOpWrapper(fn, op_type_name)
            setattr(module, fn_name, tf_op_wrapper)

            if hasattr(module, op_type_name):
                setattr(module, op_type_name, tf_op_wrapper)

            # Wraps `fn` from the public API
            if hasattr(fn, "_tf_api_names"):
                tf_api_names = getattr(fn, "_tf_api_names")
                for api_name in tf_api_names:
                    items = api_name.split(".")
                    module_names = items[:-1]
                    name = items[-1]

                    curr_module = tf
                    for curr_name in module_names:
                        curr_module = getattr(curr_module, curr_name)
                    setattr(curr_module, name, tf_op_wrapper)

            # TODO(andrey-churkin): This is required because `tensorflow.python.ops.math_ops` module
            # contains following import `from tensorflow.python.ops.gen_math_ops import *`
            if getattr(math_ops, fn_name, None) is fn:
                setattr(math_ops, fn_name, tf_op_wrapper)

        ops.name_scope = TFPatcher._wrap_name_scope_internal_fn(ops.name_scope)
        ops.name_scope_v2.__enter__ = TFPatcher._wrap_name_scope_v2_enter_fn(ops.name_scope_v2.__enter__)
        tf.name_scope.__enter__ = TFPatcher._wrap_name_scope_v2_enter_fn(tf.name_scope.__enter__)

    @staticmethod
    def _get_ops_info(op_type_names: Optional[List[str]] = None):
        raw_ops = inspect.getmembers(tf.raw_ops, predicate=inspect.isfunction)
        op_type_name_to_fn_map = dict(raw_ops)

        if op_type_names is None:
            op_type_names = list(op_type_name_to_fn_map)

        op_type_name_to_op_info_map = {}
        for op_type_name in op_type_names:
            original_fn = op_type_name_to_fn_map[op_type_name]

            module = inspect.getmodule(original_fn)
            fn_name = original_fn.__name__
            fn = getattr(module, fn_name)

            op_type_name_to_op_info_map[op_type_name] = (fn, fn_name, module)

        return op_type_name_to_op_info_map

    @staticmethod
    def _wrap_name_scope_internal_fn(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if len(args) == 4:
                args = replace_value_by_index(args, 3, False)
            else:
                kwargs["skip_on_eager"] = False
            return func(*args, **kwargs)

        return wrapper

    @staticmethod
    def _wrap_name_scope_v2_enter_fn(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            tracing_context = get_current_context()

            if tf.executing_eagerly() and tracing_context.in_call:
                (obj,) = args  # self
                eager_context = context.context()
                old_name = eager_context.scope_name
                name = obj._name

                if not name:
                    scope_name = ""
                elif name[-1] == "/":
                    scope_name = name
                elif old_name:
                    scope_name = tracing_context.unique_name(old_name + name) + "/"
                else:
                    scope_name = name + "/"
                eager_context.scope_name = scope_name

                def _restore_name_scope(*_):
                    eager_context.scope_name = old_name

                obj._exit_fns.append(_restore_name_scope)
            else:
                scope_name = func(*args, **kwargs)
            return scope_name

        return wrapper


def patch_tf_operations() -> None:
    """
    Applies patches to the TensorFlow code.
    """
    TFPatcher.patch_tf_operations()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.deprecation import warning_deprecated

warning_deprecated(
    "Importing from nncf.common.utils.logger is deprecated. "
    "Import `from nncf` directly instead, i.e.: \n"
    "`from nncf import set_log_level` instead of `from nncf.common.utils.logger import set_log_level`, and:\n"
    "`from nncf import nncf_logger` instead of `from nncf.common.utils.logger import logger as nncf_logger`"
)


from nncf.common.logging.logger import disable_logging as disable_logging  # noqa: E402
from nncf.common.logging.logger import nncf_logger as logger  # noqa: E402, F401
from nncf.common.logging.logger import set_log_level as set_log_level  # noqa: E402



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import nncf
from nncf.common.logging import nncf_logger
from nncf.common.quantization.initialization.range import PerLayerRangeInitConfig
from nncf.common.quantization.initialization.range import RangeInitConfig
from nncf.config.config import NNCFConfig
from nncf.config.schemata.defaults import NUM_BN_ADAPTATION_SAMPLES
from nncf.config.structures import BNAdaptationInitArgs
from nncf.config.structures import QuantizationRangeInitArgs


def extract_algorithm_names(config: NNCFConfig) -> List[str]:
    retval = []
    compression_config_json_section = config.get("compression", [])
    if isinstance(compression_config_json_section, dict):
        compression_config_json_section = [compression_config_json_section]
    for algo_config in compression_config_json_section:
        retval.append(algo_config["algorithm"])
    return retval


def extract_algo_specific_config(config: NNCFConfig, algo_name_to_match: str) -> Dict[str, Any]:
    """
    Extracts a .json sub-dictionary for a given compression algorithm from the
    common NNCFConfig.

    :param config: An instance of the NNCFConfig.
    :param algo_name_to_match: The name of the algorithm for which the algorithm-specific section
      should be extracted.
    :return: The sub-dictionary, exactly as it is specified in the NNCF configuration of the .json file,
    that corresponds to the algorithm-specific data (i.e. {"algorithm": "quantization", ... })
    """
    compression_section = config.get("compression", [])
    if isinstance(compression_section, list):
        algo_list = compression_section
    else:
        assert isinstance(compression_section, dict)
        algo_list = [compression_section]

    from nncf.common.compression import NO_COMPRESSION_ALGORITHM_NAME

    if algo_name_to_match == NO_COMPRESSION_ALGORITHM_NAME:
        if len(algo_list) > 0:
            raise nncf.ValidationError(
                f"No algorithm configuration should be specified "
                f"when you try to extract {algo_name_to_match} from the NNCF config!"
            )
        return {}

    matches = []
    for compression_algo_dict in algo_list:
        algo_name = compression_algo_dict["algorithm"]
        if algo_name == algo_name_to_match:
            matches.append(compression_algo_dict)

    if len(matches) > 1:
        raise nncf.ValidationError(
            f"Multiple algorithm configurations specified for the same "
            f"algo {algo_name_to_match} in the NNCF config!"
        )
    if not matches:
        raise nncf.InternalError(
            f"Did not find an algorithm configuration for algo {algo_name_to_match} in the NNCF config!"
        )
    return next(iter(matches))


def extract_range_init_params(config: NNCFConfig, algorithm_name: str = "quantization") -> Optional[Dict[str, Any]]:
    """
    Extracts parameters of the quantization range initialization algorithm from the
    compression algorithm NNCFconfig.

    :param config: An instance of the NNCFConfig.
    :param algorithm_name: Name of the compression algorithm. Should be
        one of the following: `quantization`, `experimental_quantization`.
    :return: Parameters of the quantization range initialization algorithm.
    """
    algo_config = extract_algo_specific_config(config, algorithm_name)
    init_range_config_dict_or_list = algo_config.get("initializer", {}).get("range", {})

    try:
        range_init_args = config.get_extra_struct(QuantizationRangeInitArgs)
    except KeyError:
        if not init_range_config_dict_or_list:
            nncf_logger.warning(
                "Initializer section not specified for quantization algorithm in NNCF config and "
                "quantization init args not supplied - the quantizer range initialization algorithm "
                "cannot proceed."
            )
            return None

    if not init_range_config_dict_or_list:
        nncf_logger.warning("Enabling quantization range initialization with default parameters.")
        init_range_config_dict_or_list = {"num_init_samples": 256}

    max_num_init_samples = 0
    global_range_init_config = None
    scope_overrides: List[PerLayerRangeInitConfig] = []
    if isinstance(init_range_config_dict_or_list, dict):
        global_range_init_config = RangeInitConfig.from_dict(init_range_config_dict_or_list)
        max_num_init_samples = global_range_init_config.num_init_samples
    else:
        for sub_init_range_config_dict in init_range_config_dict_or_list:
            scope_overrides.append(PerLayerRangeInitConfig.from_dict(sub_init_range_config_dict))
            max_num_init_samples_config = max(scope_overrides, key=lambda x: x.num_init_samples)
            max_num_init_samples = max_num_init_samples_config.num_init_samples

    if max_num_init_samples == 0:
        return None
    if not isinstance(range_init_args, QuantizationRangeInitArgs):
        raise ValueError(
            "Should run range initialization as specified via config,"
            "but the initializing data loader is not provided as an extra struct. "
            "Refer to `NNCFConfig.register_extra_structs` and the `QuantizationRangeInitArgs` class"
        )

    params = {
        "init_range_data_loader": range_init_args.data_loader,
        "device": range_init_args.device,
        "global_init_config": global_range_init_config,
        "per_layer_range_init_configs": scope_overrides,
    }

    return params


def extract_bn_adaptation_init_params(config: NNCFConfig, algo_name: str) -> Optional[Dict[str, object]]:
    """
    Extracts parameters for initialization of an object of the class `BatchnormAdaptationAlgorithm`
    from the compression algorithm NNCFconfig.

    :param config: An instance of the NNCFConfig.
    :param algo_name: The name of the algorithm for which the params have to be extracted.
    :return: Parameters for initialization of an object of the class `BatchnormAdaptationAlgorithm` specific
      to the supplied algorithm, or None if the config specified not to perform any batchnorm adaptation.
    """
    algo_config = extract_algo_specific_config(config, algo_name)
    params = algo_config.get("initializer", {}).get("batchnorm_adaptation", {})
    return get_bn_adapt_algo_kwargs(config, params)


def has_bn_section(config: NNCFConfig, algo_name: str) -> bool:
    algo_config = extract_algo_specific_config(config, algo_name)
    return algo_config.get("initializer", {}).get("batchnorm_adaptation") is not None


class BNAdaptDataLoaderNotFoundError(RuntimeError):
    pass


def get_bn_adapt_algo_kwargs(nncf_config: NNCFConfig, params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    num_bn_adaptation_samples = params.get("num_bn_adaptation_samples", NUM_BN_ADAPTATION_SAMPLES)

    if num_bn_adaptation_samples == 0:
        return None

    try:
        args = nncf_config.get_extra_struct(BNAdaptationInitArgs)
    except KeyError:
        raise BNAdaptDataLoaderNotFoundError(
            "Unable to create the batch-norm statistics adaptation algorithm "
            "because the data loader is not provided as an extra struct. Refer to the "
            "`NNCFConfig.register_extra_structs` method and the `BNAdaptationInitArgs` class."
        ) from None

    if not isinstance(args, BNAdaptationInitArgs):
        raise BNAdaptDataLoaderNotFoundError(
            "The extra struct for batch-norm adaptation must be an instance of the BNAdaptationInitArgs class."
        )
    params = {
        "num_bn_adaptation_samples": num_bn_adaptation_samples,
        "data_loader": args.data_loader,
        "device": args.device,
    }
    return params


def extract_accuracy_aware_training_params(config: NNCFConfig) -> Dict[str, Any]:
    """
    Extracts accuracy aware training parameters from NNCFConfig.

    :param: config: An instance of the NNCFConfig.
    :return: Accuracy aware training parameters.
    """

    class NNCFAlgorithmNames:
        QUANTIZATION = "quantization"
        FILTER_PRUNING = "filter_pruning"
        SPARSITY = ["rb_sparsity", "magnitude_sparsity", "const_sparsity"]

    def validate_accuracy_aware_schema(config: NNCFConfig, params: Dict[str, Any]) -> None:
        from nncf.common.accuracy_aware_training import AccuracyAwareTrainingMode

        if params["mode"] == AccuracyAwareTrainingMode.EARLY_EXIT:
            return
        if params["mode"] == AccuracyAwareTrainingMode.ADAPTIVE_COMPRESSION_LEVEL:
            algorithms = extract_algorithm_names(config)
            if NNCFAlgorithmNames.FILTER_PRUNING in algorithms and any(
                algo in NNCFAlgorithmNames.SPARSITY for algo in algorithms
            ):
                raise nncf.ValidationError(
                    "adaptive_compression_level mode supports filter_pruning or sparsity algorithms"
                    "separately. Please, choose only one algorithm with adaptive compression level. "
                    "Take a note that you still can use it combined with quantization."
                )
            if len(algorithms) == 1 and algorithms[0] == NNCFAlgorithmNames.QUANTIZATION:
                raise nncf.ValidationError("adaptive_compression_level mode doesn't support quantization")

    accuracy_aware_training_config = config.get("accuracy_aware_training", None)

    mode = accuracy_aware_training_config.get("mode")
    params = {"mode": mode}

    if accuracy_aware_training_config.get("params") is not None:
        for param_key, param_val in accuracy_aware_training_config.get("params").items():
            params[param_key] = param_val

    validate_accuracy_aware_schema(config, params)

    return params


def has_input_info_field(config: NNCFConfig) -> bool:
    return config.get("input_info") is not None



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import TypeVar

import nncf
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.telemetry.decorator import tracked_function
from nncf.telemetry.events import MODEL_BASED_CATEGORY
from nncf.telemetry.extractors import FunctionCallTelemetryExtractor

TModel = TypeVar("TModel")


@api(canonical_alias="nncf.strip")
@tracked_function(category=MODEL_BASED_CATEGORY, extractors=[FunctionCallTelemetryExtractor("nncf.strip")])
def strip(model: TModel, do_copy: bool = True) -> TModel:
    """
    Returns the model object with as much custom NNCF additions as possible removed
    while still preserving the functioning of the model object as a compressed model.

    :param model: The compressed model.
    :param do_copy: If True (default), will return a copy of the currently associated model object. If False,
      will return the currently associated model object "stripped" in-place.
    :return: The stripped model.
    """
    model_backend = get_backend(model)
    if model_backend == BackendType.TORCH:
        from nncf.torch.strip import strip as strip_pt

        return strip_pt(model, do_copy)  # type: ignore
    elif model_backend == BackendType.TENSORFLOW:
        from nncf.tensorflow.strip import strip as strip_tf

        return strip_tf(model, do_copy)  # type: ignore

    raise nncf.UnsupportedBackendError(f"Method `strip` does not support for {model_backend.value} backend.")



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, TypeVar

from nncf import NNCFConfig
from nncf.common.compression import BaseCompressionAlgorithmBuilder
from nncf.tensorflow.graph.model_transformer import TFModelTransformer

TModel = TypeVar("TModel")


class TFCompressionAlgorithmBuilder(BaseCompressionAlgorithmBuilder):
    """
    Determines which modifications should be made to the original model in
    order to enable algorithm-specific compression during fine-tuning.
    """

    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)
        compression_lr_multiplier = config.get_redefinable_global_param_value_for_algo(
            "compression_lr_multiplier", self.name
        )
        if compression_lr_multiplier is not None:
            raise Exception(
                "compression_lr_multiplier is not supported when your work with a TF model in NNCF. "
                "Please remove the compression_lr_multiplier attribute from your NNCFConfig."
            )

    def _get_state_without_name(self) -> Dict[str, Any]:
        """
        Implementation of get_state that returns state without builder name.

        :return: Returns a dictionary with Python data structures
            (dict, list, tuple, str, int, float, True, False, None) that represents state of the object.
        """
        return {}

    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        """
        Implementation of load state that takes state without builder name.

        :param state_without_name: Output of `_get_state_without_name()` method.
        """

    def apply_to(self, model: TModel) -> TModel:
        """
        Applies algorithm-specific modifications to the model.

        :param model: The original uncompressed model.
        :return: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        """
        transformation_layout = self.get_transformation_layout(model)
        transformer = TFModelTransformer(model)
        transformed_model = transformer.transform(transformation_layout)

        if self.should_init:
            self.initialize(transformed_model)

        return transformed_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Type, TypeVar

import torch
import torch.nn as nn

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import CONST_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.data import Dataset
from nncf.experimental.torch.sparsify_activations.sparsify_activations_impl import SparsifyActivationsAlgoBackend
from nncf.tensor.functions.torch_numeric import quantile
from nncf.torch.graph import operator_metatypes as om
from nncf.torch.graph.transformations.commands import PTSharedFnInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.model_transformer import PTModelTransformer
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.utils import training_mode_switcher

ACTIVATIONS_SPARSIFIER_PREFIX = "activations_sparsifier"
TModel = TypeVar("TModel")


class ActivationsSparsifier(nn.Module):
    """
    Sparsifies input activations by masking out values around zero.
    """

    def __init__(self, target_sparsity: float, alpha: float = 0.2):
        """
        :param target_sparsity: The target activation sparsity level.
        :param alpha: The exponential moving average decay factor in range (0, 1) for calibrating
            the threshold. A larger alpha will give more weight to the most recent batches.
        """
        super().__init__()
        self.target_sparsity = target_sparsity
        if alpha <= 0.0 or alpha >= 1.0:
            raise ValueError("The decay factor `alpha` should be in range (0, 1).")
        self.alpha = alpha
        self.register_buffer("running_threshold", torch.tensor(float("-inf")))
        self.register_buffer("num_batches_tracked", torch.tensor(0))
        self.running_threshold: torch.Tensor
        self.num_batches_tracked: torch.Tensor
        self._freeze = True

    @staticmethod
    def calculate_threshold(x: torch.Tensor, target_sparsity: float) -> torch.Tensor:
        """
        Calculates the threshold to sparsify the input tensor with target sparsity if locations of
        `x.abs() <= threshold` are zeroed out.

        :param x: The input tensor.
        :param target_sparsity: The target sparsity level on the input tensor.
        :return: The threshold value.
        """
        return quantile(x.detach().abs().view(-1), q=target_sparsity, axis=0)

    @property
    def freeze(self):
        return self._freeze

    @freeze.setter
    def freeze(self, value: bool):
        self._freeze = value

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.freeze:
            threshold = self.calculate_threshold(x, self.target_sparsity)
            self._update(threshold, dtype=x.dtype)
        mask = torch.le(x.abs(), self.running_threshold)
        x = torch.masked_fill(x, mask, 0.0)
        return x

    def reset_running_stats(self):
        """
        Resets the running threshold and the number of tracked batches to the initial stage.
        """
        self.running_threshold.fill_(float("-inf"))
        self.num_batches_tracked.zero_()

    def extra_repr(self) -> str:
        return f"target_sparsity={self.target_sparsity}"

    def _update(self, threshold: torch.Tensor, dtype: torch.dtype) -> torch.Tensor:
        """
        Updates the running threshold by exponential moving average with decaying adjustment.
        The updating logic is similar to `pandas.DataFrame.ewm(adjust=True)`.

        :param threshold: The threshold value derived from this batch to update the running threshold.
        :param dtype: Data type of the updated running threshold.
        :return: The updated running threshold.
        """
        if self.num_batches_tracked == 0:
            running_threshold = threshold
        else:
            beta = 1.0 - self.alpha
            old_running_threshold = self.running_threshold.to(device=threshold.device, dtype=torch.float64)
            running_threshold = (
                threshold.to(torch.float64) * self.alpha
                + old_running_threshold * beta * (1 - beta**self.num_batches_tracked)
            ) / (1 - beta ** (self.num_batches_tracked + 1))
        self.running_threshold = running_threshold.type(dtype)
        self.num_batches_tracked += 1
        return self.running_threshold


class PTSparsifyActivationsAlgoBackend(SparsifyActivationsAlgoBackend):
    """
    Torch backend for the activation sparsification algorithm.
    """

    SUPPORTED_METATYPES = [om.PTLinearMetatype]

    @staticmethod
    def get_sparsifiers(model: NNCFNetwork) -> List[ActivationsSparsifier]:
        """
        Finds all the activation sparsifiers in the model.

        :param model: The model with activation sparsifiers.
        :return: List of activation sparsifiers.
        """
        return [m for m in model.nncf.modules() if isinstance(m, ActivationsSparsifier)]

    @property
    def supported_metatypes(self) -> List[Type[OperatorMetatype]]:
        return PTSparsifyActivationsAlgoBackend.SUPPORTED_METATYPES

    def insert_sparsifiers(
        self,
        model: NNCFNetwork,
        graph: NNCFGraph,
        target_sparsity_by_node: Dict[NNCFNode, float],
    ) -> NNCFNetwork:
        transformation_layout = PTTransformationLayout()
        for node, target_sparsity in target_sparsity_by_node.items():
            activation_port_id = self._get_activation_port_id(node, graph)
            sparsifier = ActivationsSparsifier(target_sparsity=target_sparsity)
            sparsifier_name = f"{ACTIVATIONS_SPARSIFIER_PREFIX}_{node.node_name.replace('.', '_')}"
            transformation_layout.register(
                PTSharedFnInsertionCommand(
                    [
                        PTTargetPoint(
                            target_type=TargetType.PRE_LAYER_OPERATION,
                            target_node_name=node.node_name,
                            input_port_id=activation_port_id,
                        )
                    ],
                    sparsifier,
                    sparsifier_name,
                )
            )

        transformed_model = PTModelTransformer(model).transform(transformation_layout)
        return transformed_model

    def calibrate_sparsifiers(self, model: NNCFNetwork, graph: NNCFGraph, dataset: Dataset) -> NNCFNetwork:
        sparsifiers = self.get_sparsifiers(model)
        for sparsifier in sparsifiers:
            sparsifier.reset_running_stats()
            sparsifier.freeze = False
        with training_mode_switcher(model, is_training=False):
            with torch.no_grad():
                self.do_inference(model, dataset)
        for sparsifier in sparsifiers:
            sparsifier.freeze = True
        return model

    @staticmethod
    def _get_activation_port_id(node: NNCFNode, graph: NNCFGraph) -> int:
        """
        Finds the input activation port id for the node.

        :param node: The node to find its activation port id.
        :param graph: The NNCF graph containing the node.
        :return: The activation port id.
        """
        activation_ports = []
        for prev_node in graph.get_previous_nodes(node):
            edge = graph.get_edge(prev_node, node)
            if prev_node.metatype in CONST_NOOP_METATYPES or edge.input_port_id in node.metatype.weight_port_ids:
                continue
            activation_ports.append(edge.input_port_id)
        if len(activation_ports) != 1:
            raise nncf.InternalError(f'Cannot find activation port for node "{node}".')
        return activation_ports[0]



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.tensorflow.layers.custom_objects import NNCF_CUSTOM_OBJECTS
from nncf.tensorflow.layers.custom_objects import NNCF_QUANTIZATION_OPERATIONS
from nncf.tensorflow.layers.operation import InputType
from nncf.tensorflow.quantization.quantizers import Quantizer
from nncf.tensorflow.quantization.quantizers import TFQuantizerSpec


@NNCF_CUSTOM_OBJECTS.register()
class FakeQuantize(tf.keras.layers.Layer):
    def __init__(self, config: TFQuantizerSpec, data_format: str = "channels_last", **kwargs):
        """
        Create a FakeQuantize layer.
        """
        super().__init__(**kwargs)
        self._mode = config.mode
        self.data_format = data_format

        self._op_name = f"{self.name}_quantizer"
        self._quantizer = self._create_quantizer(config, self._op_name)
        self._quantizer_weights = {}

    @property
    def num_bits(self):
        return getattr(self._quantizer, "num_bits", None)

    @property
    def per_channel(self):
        return getattr(self._quantizer, "per_channel", None)

    @property
    def narrow_range(self):
        return getattr(self._quantizer, "narrow_range", None)

    @property
    def signed(self) -> bool:
        """
        Returns `True` for signed quantization, `False` for unsigned.

        :return: `True` for signed quantization, `False` for unsigned.
        """
        if self._quantizer.mode == QuantizationMode.SYMMETRIC:
            return self._quantizer.signed(self._quantizer_weights)
        return True

    @property
    def mode(self) -> str:
        """
        Returns mode of the quantization (symmetric or asymmetric).

        :return: The mode of the quantization.
        """
        return self._mode

    @property
    def op_name(self):
        return self._op_name

    @property
    def enabled(self):
        return self._quantizer.enabled

    @enabled.setter
    def enabled(self, v):
        self._quantizer.enabled = v

    def build(self, input_shape):
        self._quantizer_weights = self._quantizer.build(input_shape, InputType.INPUTS, self.name, self)

    def call(self, inputs, training=None):
        training = self._get_training_value(training)
        return self._quantizer(inputs, self._quantizer_weights, training)

    def register_hook_pre_quantizer(self, hook):
        return self._quantizer.register_hook_pre_call(hook)

    def apply_range_initialization(self, min_values, max_values, min_range=0.1, eps=0.01):
        self._quantizer.apply_range_initialization(self._quantizer_weights, min_values, max_values, min_range, eps)

    def _create_quantizer(self, qspec: TFQuantizerSpec, op_name: str) -> Quantizer:
        quantizer_cls = NNCF_QUANTIZATION_OPERATIONS.get(qspec.mode)
        return quantizer_cls(op_name, qspec)

    @staticmethod
    def _get_training_value(training):
        if training is None:
            training = tf.keras.backend.learning_phase()
            if tf.is_tensor(training):
                training = tf.cast(training, tf.bool)
            else:
                training = bool(training)
        return training

    def get_config(self):
        config = super().get_config()
        config.update(
            {"quantizer_config": {**self._quantizer.get_config()["quantizer_spec"]}, "data_format": self.data_format}
        )
        return config

    @classmethod
    def from_config(cls, config):
        config = config.copy()
        quantizer_config = config.pop("quantizer_config")
        return cls(TFQuantizerSpec(**quantizer_config), **config)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

import tensorflow as tf

from nncf.common.pruning.tensor_processor import NNCFPruningBaseTensorProcessor
from nncf.common.tensor import NNCFTensor
from nncf.tensorflow.tensor import TFNNCFTensor


class TFNNCFPruningTensorProcessor(NNCFPruningBaseTensorProcessor):
    """
    A realization of the processing methods set for TFNNCFTensors.
    """

    @classmethod
    def concatenate(cls, tensors: List[NNCFTensor], axis: int) -> NNCFTensor:
        ret_tensor = tf.concat([t.tensor for t in tensors], axis=axis)
        return TFNNCFTensor(ret_tensor)

    @classmethod
    def ones(cls, shape: Union[int, List[int]], device: tf.device) -> NNCFTensor:
        with tf.device(device):
            return TFNNCFTensor(tf.ones(shape))

    @classmethod
    def assert_allclose(cls, tensors: List[NNCFTensor]) -> None:
        for input_mask in tensors[1:]:
            tf.debugging.assert_near(tensors[0].tensor, input_mask.tensor)

    @classmethod
    def repeat(cls, tensor: NNCFTensor, repeats: int) -> NNCFTensor:
        ret_tensor = tf.repeat(tensor.tensor, repeats=repeats)
        return TFNNCFTensor(ret_tensor)

    @classmethod
    def elementwise_mask_propagation(cls, input_masks: List[NNCFTensor]) -> NNCFTensor:
        cls.assert_allclose(input_masks)
        return input_masks[0]

    @classmethod
    def split(cls, tensor: NNCFTensor, output_shapes: List[int]) -> List[NNCFTensor]:
        chunks = len(output_shapes)
        ret_tensors = tf.split(tensor.tensor, chunks)
        return [TFNNCFTensor(ret_tensor) for ret_tensor in ret_tensors]
